import warnings

import numpy as np
from scipy.integrate import IntegrationWarning, quad, quad_vec

# Load the C library
import os.path
from pathlib import Path
import ctypes
# # Commands to manually generate
# gcc -Wall -fPIC -c voigt.c
# gcc -shared -o libvoigt.so voigt.o
dllabspath = Path(os.path.dirname(os.path.abspath(__file__)))  # Path to libraries directory
try:
    libfile = [str(i) for i in dllabspath.glob('ext_voigtlib.*.so')][0]  # Select first (and only) library
    lib = ctypes.CDLL(libfile)  # Load the library
    lib.func.restype = ctypes.c_double  # Specify the expected result type
    lib.func.argtypes = (ctypes.c_int, ctypes.c_double)  # Specify the type of the input parameters
    cvoigt = lib.func  # Create alias for the specific function used in functions below
except IndexError:  # File does not exist
    warnings.warn("Could not locate the external C library. Further use of `clib` will fail!")


# Parameters for `voigt_approx_nobg` and other approx. Voigt functions
params = np.array([[-1.2150, -1.3509, -1.2150, -1.3509],
                   [1.2359, 0.3786, -1.2359, -0.3786],
                   [-0.3085, 0.5906, -0.3085, 0.5906],
                   [0.0210, -1.1858, -0.0210, 1.1858]])
sqrt_ln2 = np.sqrt(np.log(2))
sqrt_pi = np.sqrt(np.pi)
A, B, C, D = params


__all__ = ['voigt_approx_nobg', 'voigt_approx', 'double_voigt_approx_nobg', 'double_voigt_approx',
           'voigt_nobg', 'voigt', 'double_voigt_nobg', 'double_voigt']


def voigt_approx_nobg(x, a, b, s, g):
    """Voigt function (efficient approximation) with no background (Base approx. Voigt function)

    This is the base for all other approximated Voigt functions. Not implemented in any models yet as initial tests
    exhibited slow convergence.

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a : float
        Amplitude of the Lorentzian.
    b : float
        Central line core.
    s : float
        Sigma (for Gaussian).
    g : float
        Gamma (for Lorentzian).

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt_approx : Approximated Voigt function with background added
    double_voigt_approx_nobg : Two approximated Voigt functions added together
    double_voigt_approx : Two approximated Voigt functions and a background added together
    voigt_nobg : Base Voigt function with no background
    voigt : Voigt function with background added
    double_voigt_nobg : Two Voigt functions added together
    double_voigt : Two Voigt function and a background added together

    Notes
    -----
    This algorithm is taken from A. B. McLean et al. [1]_.

    References
    ----------
    .. [1] A. B. McLean, C. E. J. Mitchell and D. M. Swanston, "Implementation of an efficient analytical
      approximation to the Voigt function for photoemission lineshape analysis," Journal of Electron Spectroscopy and
      Related Phenomena, vol. 69, pp. 125-132, 1994. https://doi.org/10.1016/0368-2048(94)02189-7
    """
    fwhm_g = 2 * s * np.sqrt(2 * np.log(2))
    fwhm_l = 2 * g
    xx = (x - b) * 2 * sqrt_ln2 / fwhm_g
    xx = xx[..., np.newaxis]
    yy = fwhm_l * sqrt_ln2 / fwhm_g
    yy = yy[..., np.newaxis]
    v = np.sum((C * (yy - A) + D * (xx - B)) / ((yy - A) ** 2 + (xx - B) ** 2), axis=-1)
    return fwhm_l * a * sqrt_pi / fwhm_g * v


def voigt_approx(x, a, b, s, g, d):
    """Voigt function (efficient approximation) with background

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a : float
        Amplitude of the Lorentzian.
    b : float
        Central line core.
    s : float
        Sigma (for Gaussian).
    g : float
        Gamma (for Lorentzian).
    d : float
        Background.

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt_approx_nobg : Base approximated Voigt function with no background
    double_voigt_approx_nobg : Two approximated Voigt functions added together
    double_voigt_approx : Two approximated Voigt functions and a background added together
    voigt_nobg : Base Voigt function with no background
    voigt : Voigt function with background added
    double_voigt_nobg : Two Voigt functions added together
    double_voigt : Two Voigt function and a background added together
    """
    return voigt_approx_nobg(x, a, b, s, g) + d


def double_voigt_approx_nobg(x, a1, b1, s1, g1, a2, b2, s2, g2):
    """Double Voigt function (efficient approximation) with no background

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a1 : float
        Amplitude of the Lorentzian of 1st Voigt function.
    b1 : float
        Central line core of 1st Voigt function.
    s1 : float
        Sigma (for Gaussian) of 1st Voigt function.
    g1 : float
        Gamma (for Lorentzian) of 1st Voigt function.
    a2 : float
        Amplitude of 2st Voigt function.
    b2 : float
        Central line core of 2st Voigt function.
    s2 : float
        Sigma (for Gaussian) of 2st Voigt function.
    g2 : float
        Gamma (for Lorentzian) of 2st Voigt function.

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt_approx_nobg : Base approximated Voigt function with no background
    voigt_approx : Approximated Voigt function with background added
    double_voigt_approx : Two approximated Voigt functions and a background added together
    voigt_nobg : Base Voigt function with no background
    voigt : Voigt function with background added
    double_voigt_nobg : Two Voigt functions added together
    double_voigt : Two Voigt function and a background added together
    """
    return voigt_approx_nobg(x, a1, b1, s1, g1) + voigt_approx_nobg(x, a2, b2, s2, g2)


def double_voigt_approx(x, a1, b1, s1, g1, a2, b2, s2, g2, d):
    """Double Voigt function (efficient approximation) with background

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a1 : float
        Amplitude of the Lorentzian of 1st Voigt function.
    b1 : float
        Central line core of 1st Voigt function.
    s1 : float
        Sigma (for Gaussian) of 1st Voigt function.
    g1 : float
        Gamma (for Lorentzian) of 1st Voigt function.
    a2 : float
        Amplitude of 2st Voigt function.
    b2 : float
        Central line core of 2st Voigt function.
    s2 : float
        Sigma (for Gaussian) of 2st Voigt function.
    g2 : float
        Gamma (for Lorentzian) of 2st Voigt function.
    d : float
        Background.

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt_approx_nobg : Base approximated Voigt function with no background
    voigt_approx : Approximated Voigt function with background added
    double_voigt_approx_nobg : Two approximated Voigt functions added together
    voigt_nobg : Base Voigt function with no background
    voigt : Voigt function with background added
    double_voigt_nobg : Two Voigt functions added together
    double_voigt : Two Voigt function and a background added together
    """
    return voigt_approx_nobg(x, a1, b1, s1, g1) + voigt_approx_nobg(x, a2, b2, s2, g2) + d


def voigt_nobg(x, a, b, s, g, clib=True):
    """Voigt function with no background (Base Voigt function)

    This is the base of all the other Voigt functions.

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a : float
        Amplitude.
    b : float
        Central line core.
    s : float
        Sigma (for Gaussian).
    g : float
        Gamma (for Lorentzian).
    clib : bool, optional, default = True
        Whether to use the complied C library or a slower Python version. If using the C library, the accuracy
        of the integration is reduced to give the code a significant speed boost. Python version can be used when
        speed is not a priority. Python version will remove deviations that are sometimes present around the wings
        due to the reduced accuracy.

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt : Voigt function with background added
    double_voigt_nobg : Two Voigt functions added together
    double_voigt : Two Voigt function and a background added together

    Notes
    -----
    More information on the Voigt function can be found here: https://en.wikipedia.org/wiki/Voigt_profile
    """
    warnings.filterwarnings("ignore", category=IntegrationWarning)
    u = x - b
    if clib:
        i = [quad(cvoigt, -np.inf, np.inf, args=(v, s, g), epsabs=1.49e-1, epsrel=1.49e-4)[0] for v in u]
    else:
        i = quad_vec(lambda y: np.exp(-y**2 / (2 * s**2)) / (g**2 + (u - y)**2), -np.inf, np.inf)[0]
    const = g / (s * np.sqrt(2 * np.pi**3))
    return a * const * np.array(i)


def voigt(x, a, b, s, g, d, clib=True):
    """Voigt function with background

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a : float
        Amplitude.
    b : float
        Central line core.
    s : float
        Sigma (for Gaussian).
    g : float
        Gamma (for Lorentzian).
    d : float
        Background.
    clib : bool, optional, default = True
        Whether to use the complied C library or a slower Python version. If using the C library, the accuracy
        of the integration is reduced to give the code a significant speed boost. Python version can be used when
        speed is not a priority. Python version will remove deviations that are sometimes present around the wings
        due to the reduced accuracy.

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt_nobg : Base Voigt function with no background
    double_voigt_nobg : Two Voigt functions added together
    double_voigt : Two Voigt function and a background added together

    Notes
    -----
    More information on the Voigt function can be found here: https://en.wikipedia.org/wiki/Voigt_profile
    """
    return voigt_nobg(x, a, b, s, g, clib) + d


def double_voigt_nobg(x, a1, b1, s1, g1, a2, b2, s2, g2, clib=True):
    """Double Voigt function with no background

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a1 : float
        Amplitude of 1st Voigt function.
    b1 : float
        Central line core of 1st Voigt function.
    s1 : float
        Sigma (for Gaussian) of 1st Voigt function.
    g1 : float
        Gamma (for Lorentzian) of 1st Voigt function.
    a2 : float
        Amplitude of 2st Voigt function.
    b2 : float
        Central line core of 2st Voigt function.
    s2 : float
        Sigma (for Gaussian) of 2st Voigt function.
    g2 : float
        Gamma (for Lorentzian) of 2st Voigt function.
    clib : bool, optional, default = True
        Whether to use the complied C library or a slower Python version. If using the C library, the accuracy
        of the integration is reduced to give the code a significant speed boost. Python version can be used when
        speed is not a priority. Python version will remove deviations that are sometimes present around the wings
        due to the reduced accuracy.

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt_nobg : Base Voigt function with no background
    voigt : Voigt function with background added
    double_voigt : Two Voigt function and a background added together

    Notes
    -----
    More information on the Voigt function can be found here: https://en.wikipedia.org/wiki/Voigt_profile
    """
    return voigt_nobg(x, a1, b1, s1, g1, clib) + voigt_nobg(x, a2, b2, s2, g2, clib)


def double_voigt(x, a1, b1, s1, g1, a2, b2, s2, g2, d, clib=True):
    """Double Voigt function with background

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Voigt function at.
    a1 : float
        Amplitude of 1st Voigt function.
    b1 : float
        Central line core of 1st Voigt function.
    s1 : float
        Sigma (for Gaussian) of 1st Voigt function.
    g1 : float
        Gamma (for Lorentzian) of 1st Voigt function.
    a2 : float
        Amplitude of 2st Voigt function.
    b2 : float
        Central line core of 2st Voigt function.
    s2 : float
        Sigma (for Gaussian) of 2st Voigt function.
    g2 : float
        Gamma (for Lorentzian) of 2st Voigt function.
    d : float
        Background.
    clib : bool, optional, default = True
        Whether to use the complied C library or a slower Python version. If using the C library, the accuracy
        of the integration is reduced to give the code a significant speed boost. Python version can be used when
        speed is not a priority. Python version will remove deviations that are sometimes present around the wings
        due to the reduced accuracy.

    Returns
    -------
    result : ndarray of shape `x.shape`
        The value of the Voigt function here.

    See Also
    --------
    voigt_nobg : Base Voigt function with no background
    voigt : Voigt function with background added
    double_voigt_nobg : Two Voigt functions added together

    Notes
    -----
    More information on the Voigt function can be found here: https://en.wikipedia.org/wiki/Voigt_profile
    """
    return double_voigt_nobg(x, a1, b1, s1, g1, a2, b2, s2, g2, clib) + d



import numpy as np


__all__ = ['genmask', 'radial_distances']


def genmask(width, height, radius=np.inf, right_shift=0, up_shift=0):
    """Generate a circular mask of specified size

    Parameters
    ----------
    width : int
        Width of mask.
    height : int
        Height of mask.
    radius : int, optional, default = inf
        Radius of mask.
    right_shift : int, optional, default = 0
        Indices to shift forward through row.
    up_shift : int, optional, default = 0
        Indices to shift forward through columns.

    Returns
    -------
    array : ndarray of shape (height, width)
        The generated mask.
    """
    array = radial_distances(width, height) < radius  # Create mask
    array = np.roll(array, [up_shift, right_shift], [0, 1])  # Apply shift
    return array


def radial_distances(n_cols, n_rows):
    """Generates a 2D array of specified shape of radial distances from the centre

    Parameters
    ----------
    n_cols : int
        Number of columns.
    n_rows : int
        Number of rows.

    Returns
    -------
    array : ndarray of shape (n_rows, n_cols)
        Array of radial distances.

    See Also
    --------
    genmask : Generates a circular mask
    """
    horiz_mid = (n_cols - 1) / 2.0  # Horizontal midpoint
    verti_mid = (n_rows - 1) / 2.0  # Vertical midpoint
    array = np.empty((n_rows, n_cols))
    for i in range(n_rows):
        for j in range(n_cols):
            array[i, j] = np.sqrt((j - horiz_mid)**2.0 + (i - verti_mid)**2.0)
    return array



import numpy as np
import scipy.ndimage


__all__ = ['moving_average', 'gaussian_kern_3d', 'smooth_cube']


def moving_average(array, width):
    """Boxcar moving average

    Calculate the moving average of an array with a boxcar of defined width. An odd width is recommended.

    Parameters
    ----------
    array : ndarray, ndim=1
        Array to find the moving average of.
    width : int
        Width of the boxcar. Odd integer recommended. Less than or equal to length of `array`.

    Returns
    -------
    averaged : ndarray of shape `array`
        Averaged array.

    Notes
    -----
    The moving average is calculated at each point of the `array` by finding the (unweighted) mean of the subarrays
    of length given by `width`. These subarrays are centred at the point in the `array` that the current average is
    currently being calculated for. If an odd `width` is chosen, the sub array will include the current point plus an
    equal number of points on either side. However, if an even `width` is chosen, the sub array will bias including the
    extra point to the left of the current index. If the subarray spans past the boundaries, the values beyond the
    boundary is ignored and the mean is calculated by dividing by the number of points that are inside the boundaries.

    Examples
    --------
    >>> x = np.array([1, 2, 3, 4, 5])
    >>> moving_average(x, 3)
    array([1.5, 2. , 3. , 4. , 4.5])
    >>> moving_average(x, 2)
    array([1. , 1.5, 2.5, 3.5, 4.5])
    """

    if not isinstance(width, int) or width <= 0 or width > len(array):
        raise ValueError("`width` must be a positive integer less than the length of `array`, got %s." % width)

    kernel = np.ones(width) / width
    averaged = np.convolve(array, kernel, mode='same')

    # Apply corrections at edge
    to_correct_left = int(width/2)
    to_correct_right = to_correct_left
    to_correct_right -= 0 if width % 2 else 1

    # Adjust weighting on left edge
    n_overlap = width - 1
    for i in np.arange(to_correct_left-1, -1, -1):
        averaged[i] *= width / n_overlap
        n_overlap -= 1

    # Adjust weighting on right edge
    n_overlap = width - 1
    for i in range(len(array) - to_correct_right, len(array)):
        averaged[i] *= width / n_overlap
        n_overlap -= 1

    return averaged


def gaussian_kern_3d(width=5, sigma=(1, 1, 1)):
    """3D Gaussian kernel

    Create a Gaussian kernel of shape `width`*`width`*`width`.

    Parameters
    ----------
    width : int, optional, default = 5
        Length of all three dimensions of the Gaussian kernel.
    sigma : array_like, tuple, optional, default = (1, 1, 1)
        Sigma values for the time, horizontal and vertical dimensions.

    Returns
    -------
    kernel : ndarray, shape (`width`, `width`, `width`)
        The generated kernel.

    Examples
    --------
    >>> gaussian_kern_3d(width=3, sigma=(2, 1, 1.5))
    array([[[0.42860385, 0.53526143, 0.42860385],
            [0.48567179, 0.60653066, 0.48567179],
            [0.42860385, 0.53526143, 0.42860385]],
           [[0.70664828, 0.8824969 , 0.70664828],
            [0.8007374 , 1.        , 0.8007374 ],
            [0.70664828, 0.8824969 , 0.70664828]],
           [[0.42860385, 0.53526143, 0.42860385],
            [0.48567179, 0.60653066, 0.48567179],
            [0.42860385, 0.53526143, 0.42860385]]])
    """
    s = np.linspace(-1, 1, width)
    x, y, z = np.meshgrid(s, s, s)
    kernel = np.exp(-x**2/(2*sigma[0]**2) - y**2/(2*sigma[1]**2) - z**2/(2*sigma[2]**2))
    return kernel


def smooth_cube(cube, mask, **kwargs):
    """Apply Gaussian smoothing to velocities

    Smooth the cube of velocities with a Gaussian kernel, applying weights at boundaries.

    Parameters
    ----------
    cube : ndarray, ndim=3
        Cube of velocities with dimensions [time, row, column].
    mask : ndarray, ndim=2
        The mask to apply to the [row, column] at every time. Points that are 0 or false will be removed.
    kwargs : optional
        Keyword arguments to pass to `gaussian_kern_3d`.

    Returns
    -------
    cube_ : ndarray, shape `cube`
        The smoothed cube.
    """
    # Masking
    cube_ = cube.copy()
    mask3d = np.empty_like(cube_)  # Mask for `cube_`
    for timestep in range(len(mask3d)):  # Copy mask throughout time
        mask3d[timestep] = mask.copy()
    cube_[np.isnan(cube_)] = 0  # Remove NaN
    cube_[mask3d == 0] = 0  # Remove masked points

    kernel = gaussian_kern_3d(**kwargs)  # Get the kernel

    # Count the number of neighbours (weighting) at each pixel
    neighbour_count = scipy.ndimage.convolve(mask3d, kernel, mode="constant")
    m = neighbour_count > 0  # Record the points that do have neighbours

    cube_ = scipy.ndimage.convolve(cube_, kernel, mode="constant")  # Apply smoothing
    cube_[m] /= neighbour_count[m]  # Normalise depending on neighbours

    cube_[mask3d == 0] = np.nan  # Restore masked pixels
    cube_[np.isnan(cube)] = np.nan  # Restore NaN values

    return cube_



import os

import numpy as np
from astropy.io import fits
from scipy.io import readsav


__all__ = ['make_iter', 'load_parameter']


def make_iter(*args):
    """Returns each inputted argument, wrapping in a list if not already iterable

    Parameters
    ----------
    *args
        Arguments to make iterable.

    Returns
    -------
    iterables
        `*args` converted to iterables.

    Examples
    --------
    >>> make_iter(1)
    [[1]]

    >>> make_iter(1, 2, 3)
    [[1], [2], [3]]

    >>> make_iter(1, [2], 3)
    [[1], [2], [3]]

    It is intended that a list of arguments be passed to the function for conversion:

    >>> make_iter(*[1, [2], 3])
    [[1], [2], [3]]

    Remember that strings are already iterable!

    >>> make_iter(*[[1, 2, 3], (4, 5, 6), "a"])
    [[1, 2, 3], (4, 5, 6), 'a']
    """
    iterables = []  # Holder to return
    for parameter in args:
        try:  # Will work if iterable
            _ = (i for i in parameter)
        except TypeError:  # Not iterable
            parameter = [parameter]  # Wrap in a list
        iterables = iterables + [parameter]  # Append to list to return
    return iterables


def load_parameter(parameter, wl=None):
    """Load parameters from file, optionally evaluating variables from strings

    Loads the parameter from string or file.

    Parameters
    ----------
    parameter : str
        Parameter to load, either string of Python list/number or filename string. Supported filename extensions are
        '.fits', '.fit', '.fts', '.csv', '.txt', '.npy', '.npz', and '.sav'. If the file does not exist, it will assume
        the string is a Python expression.
    wl : float, optional, default = None
        Central line core wavelength to replace 'wl' in strings. Will only replace occurrences in the `parameter`
        variable itself or in files with extension ".csv" or ".txt". When using `wl`, also use 'inf' and 'nan' as
        required.

    Returns
    -------
    value : ndarray or list of floats
        Value of parameter in easily computable format (not string)

    Examples
    --------
    >>> load_parameter("wl + 4.2", wl=7.1)
    11.3

    >>> load_parameter("[wl + 4.2, 5.2 - inf, 5 > 3]", wl=7.1)
    [11.3, -inf, 1.0]

    Filenames are given as follows:

    >>> x = load_parameter("datafile.csv", wl=12.4)

    >>> x = load_parameter("datafile.fits")

    If the file does not exist, the function will assume that the string is a Python expression, possibly leading to an
    error:

    >>> load_parameter("nonexistant.csv")
    TypeError: 'NoneType' object is not subscriptable
    """
    if os.path.exists(parameter):  # If the parameter is a real file

        ext = os.path.splitext(parameter)[1]  # File extension

        if ext.lower() in ['.fits', '.fit', '.fts']:  # Extension suggests FITS file

            # Read data from the primary HDU of the FITS file
            hdul = fits.open(parameter)  # Open with mmap
            value = hdul[0].data.copy()  # Copy out of mmap
            hdul.close()  # Close the file

        elif ext.lower() in ['.csv', '.txt']:  # Extension suggests CSV file

            # Read CSV file (assumes a ',' delimiter)
            if wl is not None:  # If `wl` is specified, try a replacement
                value = str(list(np.loadtxt(parameter, delimiter=',', dtype=object))).replace('\'', '')
                try:
                    value = eval(str(value), {'__builtins__': None},
                                 {'wl': wl, 'inf': float('inf'), 'nan': float('nan')})
                except TypeError:  # Only allowed to process `wl` and `inf` variables for security reasons
                    raise SyntaxError("parameter string contains illegal variables")
                except SyntaxError:
                    raise SyntaxError("parameter string '{}' contains a syntax error".format(parameter))
            else:
                value = np.loadtxt(parameter, delimiter=',', dtype=float)

        elif ext.lower() in ['.npy', '.npz']:  # Extension suggests NumPy array

            value = np.load(parameter)

        elif ext.lower() in ['.sav']:  # Extension suggests IDL SAVE file (assumes relevant data in first variable)

            value = list(readsav(parameter).values())[0]

        else:  # Extension not matched
            raise ValueError("loaded parameters can only have file extensions: '.fits', '.fit', '.fts', '.csv', "
                             "'.txt', '.npy', '.npz', '.sav', got '%s'" % ext.lower())

    else:  # Must not be a file (or the filename is incorrect!)

        # Convert to list, calculate relative to central line core (`wl`)
        value = eval(str(parameter), {'__builtins__': None}, {'wl': wl, 'inf': float('inf')})
        try:
            value = [float(val) for val in value]  # Make sure all values are floats
        except TypeError:
            value = float(value)

    return value



import numpy as np
from astropy.io import fits

from mcalf.utils.misc import make_iter


__all__ = ['FitResult', 'FitResults']


class FitResult:
    """Class that holds the result of a fit

    Parameters
    ----------
    fitted_parameters : ndarray
        The parameters fitted.
    fit_info : dict
        Additional information on the fit including at least 'classification', 'profile', 'success' and 'index'.

    Attributes
    ----------
    parameters : ndarray
        The parameters fitted.
    classification : int
        Classification of the fitted spectrum.
    profile : str
        Profile of the fitted spectrum.
    success : bool
        Whether the fit was completed successfully.
    index : list
        Index ([<time>, <row>, <column>]) of the spectrum in the spectral array.
    __dict__
        Other attributes may be present depending on the `fit_info` used.
    """
    def __init__(self, fitted_parameters, fit_info):
        self.__dict__ = fit_info  # Load first
        self.parameters = fitted_parameters

    def __len__(self):  # Calling the python `len` function on this object will return the number of fitted parameters
        return len(self.parameters)

    def __repr__(self):  # Useful string output of the object
        success = 'Successful ' if self.__dict__['success'] else 'Unsuccessful '
        index = ''
        if 'index' in self.__dict__:
            i = self.__dict__['index']
            if (isinstance(i, list) or isinstance(i, tuple)) and len(i) == 3 and all([j is not None for j in i]):
                index = 'at ({}, {}, {}) '.format(*i)
        return success + 'FitResult ' + index + 'with ' + self.__dict__['profile'] \
            + ' profile of classification ' + str(self.__dict__['classification'])

    def plot(self, model, **kwargs):
        """Plot the data and fitted parameters

        This calls the `plot` method on `model` but will plot for this FitResult object. See the model's `plot` method
        for more details.

        Parameters
        ----------
        model : child class of ModelBase
            The model object to plot with.
        **kwargs
            See the `model.plot` method for more details.
        """
        model.plot(self, **kwargs)

    def velocity(self, model, vtype='quiescent'):
        """Calculate the Doppler velocity of the fit using `model` parameters

        Parameters
        ----------
        model : child class of ModelBase
            The model object to take parameters from.
        vtype : {'quiescent', 'active'}, default='quiescent'
            The velocity type to find.

        Returns
        -------
        velocity : float
            The calculated velocity.
        """
        stationary_line_core = model.stationary_line_core

        if vtype == 'quiescent':
            index = model.quiescent_wavelength
        elif vtype == 'active':
            index = model.active_wavelength
        else:
            raise ValueError("unknown velocity type '%s'" % vtype)

        try:
            wavelength = self.parameters[index]  # Choose the shifted wavelength from the fitted parameters
        except IndexError:  # Fit not compatible with this velocity type
            wavelength = np.nan  # No emission fitted

        return (wavelength - stationary_line_core) / stationary_line_core * 300000  # km/s


class FitResults:
    """Class that holds multiple fit results in a way that can be easily processed

    Attributes
    ----------
    parameters : ndarray of shape (`row`, `column`, `parameter`)
        Array of fitted parameters.
    classifications : ndarray of int of shape (`row`, `column`)
        Array of classifications.
    profile : ndarray of str of shape (`row`, `column`)
        Array of profiles.
    success : ndarray of bool of shape (`row`, `column`)
        Array of success statuses.
    chi2 : ndarray of shape (`row`, `column`)
        Array of chi-squared values.
    time : int, default = None
        Time index that the `FitResult` object refers to (if provided).
    n_parameters : int
        Number of parameters in the last dimension of `parameters`.
    """
    def __init__(self, shape, n_parameters, time=None):
        """Initialise a `FitResults` object of a defined shape

        Parameters
        ----------
        shape : tuple of int
            The number of rows and columns to hold data for, e.g. (n_rows, n_columns).
        n_parameters : int
            The number of fitted parameters per spectrum that need to be stored.
        time : int, optional, default = None
            The time the `FitResults` object will store data for. Optional, but if it is set, only `FitResult` objects
            with a matching time can be appended.
        """
        # TODO Allow multiple time indices to be imported
        if not isinstance(shape, tuple) or len(shape) != 2:
            raise TypeError("`shape` must be a tuple of length 2, got %s" % type(shape))

        if not isinstance(n_parameters, int) or n_parameters < 1:
            raise ValueError("`n_parameters` must be an integer greater than zero, got %s" % n_parameters)
        parameters_shape = tuple(list(shape) + [n_parameters])
        self.parameters = np.full(parameters_shape, np.nan, dtype=float)

        self.classifications = np.full(shape, -1, dtype=int)

        self.profile = np.full(shape, '', dtype=object)

        self.success = np.full(shape, False, dtype=bool)

        self.chi2 = np.full(shape, np.nan, dtype=float)

        self.time = time
        self.n_parameters = n_parameters

    def append(self, result):
        """Append a `FitResult` object to the `FitResults` object

        Parameters
        ----------
        result : FitResult object
            `FitResult` object to append.
        """
        time, row, column = result.index
        if self.time is not None and self.time != time:
            raise ValueError("The time index of `result` does not match the time index being filled.")

        # TODO Make the number of parameters and types of profiles general.
        p = result.profile
        if p == 'absorption':
            self.parameters[row, column, :4] = result.parameters
        elif p == 'emission':
            self.parameters[row, column, 4:] = result.parameters
        elif p == 'both':
            self.parameters[row, column] = result.parameters
        else:
            raise ValueError("Unknown profile '%s'" % p)

        self.classifications[row, column] = result.classification
        self.profile[row, column] = result.profile
        self.success[row, column] = result.success
        self.chi2[row, column] = result.chi2

    def velocities(self, model, row=None, column=None, vtype='quiescent'):
        """Calculate the Doppler velocities of the fit results using `model` parameters

        Parameters
        ----------
        model : child class of ModelBase
            The model object to take parameters from.
        row : int, list, array_like, iterable, optional, default = None
            The row indices to find velocities for. All if omitted.
        column : int, list, array_like, iterable, optional, default = None
            The column indices to find velocities for. All if omitted.
        vtype : {'quiescent', 'active'}, default='quiescent'
            The velocity type to find.

        Returns
        -------
        velocities : ndarray of shape (`row`, `column`)
            The calculated velocities for the specified `row` and `column` positions.
        """
        if row is None:
            row = range(len(self.parameters))
        if column is None:
            column = range(len(self.parameters[0]))

        if vtype == 'quiescent':
            index = model.quiescent_wavelength
        elif vtype == 'active':
            index = model.active_wavelength
        else:
            raise ValueError("unknown velocity type '%s'" % vtype)

        index, row, column = make_iter(index, row, column)
        wavelengths = self.parameters[row][:, column][:, :, index]
        stationary_line_core = model.stationary_line_core
        return np.squeeze((wavelengths - stationary_line_core) / stationary_line_core * 300000, axis=2)  # km/s

    def save(self, filename, model=None):
        """Saves the FitResults object to a FITS file

        Parameters
        ----------
        filename : file path, file object or file-like object
            FITS file to write to. If a file object, must be opened in a writeable mode.
        model : child class of mcalf.models.base.ModelBase, optional, default = None
            If provided, use this model to calculate and include both quiescent and active Doppler velocities.

        Notes
        -----
        Saves a FITS file to the location specified by `filename`. All the parameters are stored in a separate,
        named, HDU.
        """
        # Compress profile array to integers
        p_uniq = np.unique(self.profile)
        p_legend = np.array_str(p_uniq)
        p = np.full_like(self.profile, -1, dtype=np.int16)
        for i in range(len(p_uniq)):
            p[self.profile == p_uniq[i]] = i

        header = fits.Header({
            'NTIME': 1,
            'NROWS': self.classifications.shape[-2],
            'NCOLS': self.classifications.shape[-1],
            'TIME': self.time,
        })
        primary_hdu = fits.PrimaryHDU([], header)

        header = fits.Header({'NPARAMS': self.n_parameters})
        parameters_hdu = fits.ImageHDU(self.parameters, header, 'PARAMETERS')

        classifications_hdu = fits.ImageHDU(np.asarray(self.classifications, dtype=np.int16), name='CLASSIFICATIONS')

        header = fits.Header({'PROFILES': p_legend})
        profile_hdu = fits.ImageHDU(p, header, 'PROFILE')

        success_hdu = fits.ImageHDU(np.asarray(self.success, dtype=np.int16), name='SUCCESS')

        chi2_hdu = fits.ImageHDU(self.chi2, name='CHI2')

        hdul = fits.HDUList([primary_hdu, parameters_hdu, classifications_hdu, profile_hdu, success_hdu, chi2_hdu])

        if model is not None:
            for head, vtype, name in [('ACTIVE', 'active', 'VLOSA'),
                                      ('QUIESCENT', 'quiescent', 'VLOSQ')]:
                header = fits.Header({'VTYPE': head, 'UNIT': 'KM/S'})
                v = self.velocities(model, vtype=vtype)
                v_hdu = fits.ImageHDU(v, header, name)
                hdul.append(v_hdu)

        hdul.writeto(filename, checksum=True)



import numpy as np
from scipy.special import erf


__all__ = ['single_gaussian', 'skew_normal', 'skew_normal_with_gaussian']


def single_gaussian(x, a, b, c, d):
    """Gaussian function

    Parameters
    ----------
    x : ndarray
        Wavelengths to evaluate Gaussian function at.
    a : float
        Amplitude.
    b : float
        Central line core.
    c : float
        Sigma of Gaussian.
    d : float
        Background to add.
    """
    g = a * np.exp(- (x-b)**2.0 / (2.0 * c**2.0))
    return g + d


def _skew_normal_with_gaussian(x, a_a=0, alpha=1, xi=0, omega=1, a_e=0, b=0, c=1, d=0):
    x1 = (x - xi) / omega
    pdf = np.exp(- x1 ** 2.0 / 2.0) / np.sqrt(2.0 * np.pi)
    cdf = 0.5 * (1.0 + erf(alpha * x1 / np.sqrt(2.0)))
    skewed_normal = 2.0 * pdf * cdf / omega

    absorption = a_a * skewed_normal
    emission = a_e * np.exp(- (x - b) ** 2.0 / (2.0 * c ** 2.0))

    return absorption + emission + d


def skew_normal(x, a_a, alpha, xi, omega, d):
    return _skew_normal_with_gaussian(x, a_a=a_a, alpha=alpha, xi=xi, omega=omega, d=d)


def skew_normal_with_gaussian(x, a_a, alpha, xi, omega, a_e, b, c, d):  # This can be optimised later
    return _skew_normal_with_gaussian(x, a_a=a_a, alpha=alpha, xi=xi, omega=omega, a_e=a_e, b=b, c=c, d=d)



import os
import warnings

import numpy as np
from scipy.optimize import curve_fit
from sklearn.exceptions import NotFittedError
from sklearn.metrics import classification_report

from mcalf.utils.spec import reinterpolate_spectrum
from mcalf.utils.misc import make_iter


__all__ = ['ModelBase']


class ModelBase:
    """Base class for spectral line model fitting.

    Warning: This class should not be used directly.
    Use derived classes instead.

    Attributes
    ----------
    array : ndarray, dimensions are ['time', 'row', 'column', 'spectra']
        Array holding spectra.
    background : ndarray, dimensions are ['time', 'row', 'column']
        Array holding spectral backgrounds.
    """
    def __init__(self):
        self.array = None  # Array that will hold any loaded spectra
        self.background = None  # Array holding constant background values for `self.array`

    def _load_data(self, array, names=None, target=None):
        """Load a specified array into the model object.

        Load `array` with dimension names `names` into the parameter specified by `target`.

        Parameters
        ----------
        array : ndarray
            The array to load.

        names : list of str, length = `array.ndims`
            List of dimension names for `array`. Valid dimension names depend on `target`.

        target : {'array', 'background'}

        See Also
        --------
        load_array : Load and array of spectra
        load_background : Load an array of spectral backgrounds
        """

        # Validate specified `target`
        if target not in ['array', 'background']:
            raise ValueError("array target must be either 'array' or 'background', got '%s'" % target)

        # Verify `array` has data on more than one spectrum
        if (target == 'array' and np.ndim(array) == 1) or (target == 'background' and np.ndim(array) == 0):
            raise ValueError("cannot load an array containing one spectrum, use the spectrum directly instead")

        # Validate the dimension names given
        if names is None:
            raise ValueError("dimension names must be specified for each dimension")
        if np.ndim(array) != len(np.atleast_1d(names)):  # All dimensions must be labelled
            raise ValueError("number of dimension names do not match number of columns")
        if len(np.atleast_1d(names)) != len(np.unique(np.atleast_1d(names))):
            raise ValueError("duplicate dimension names found")
        if target == 'array' and 'wavelength' not in names:
            raise ValueError("array must contain a wavelength dimension")

        # # Pad input array to include all possible input dimensions and
        # #     then transpose it such that order of dimensions is consistent
        #   e.g. ['time', 'wavelength']           -> ['time', 'row', 'column', 'wavelength']
        #        ['wavelength', 'column', 'row']  -> ['time', 'row', 'column', 'wavelength']

        # Define the transposition on the input array to get the dimensions in order
        #     ['time', 'row', 'column'(, 'wavelength')]
        n_dims = 4 if target == 'array' else 3  # ndims of (padded) array to load data into
        transposition = [None] * n_dims
        for i in range(len(names)):
            if names[i] == 'time':
                transposition[0] = i
            elif names[i] == 'row':
                transposition[1] = i
            elif names[i] == 'column':
                transposition[2] = i
            elif names[i] == 'wavelength' and target == 'array':  # Only `target` 'array' should contain wavelengths
                transposition[3] = i
            else:  # Name given is not recognised
                raise ValueError("name '{}' is not a valid dimension name".format(names[i]))

        # Define where the new padded dimensions will be placed within the transposition
        array_n_dims = np.ndim(array)  # Number of dimensions in input array
        new_axis_id = array_n_dims  # Start labelling new axes after last dimension of input array
        for axis in range(n_dims):  # Loop through each axis of padded array to be created
            if transposition[axis] is None:  # Axis hasn't been labelled yet (not part of input array)
                transposition[axis] = new_axis_id  # Locate it within the transposition definition
                new_axis_id += 1

        # Add missing dimensions to the end
        n_dims_to_add = n_dims - array_n_dims
        ordered_array = array
        for i in range(n_dims_to_add):  # Add required number of dimensions to the end of the input array
            ordered_array = ordered_array[..., np.newaxis]

        # Reorder the dimensions of the padded array such that they can be later assumed
        ordered_array = np.transpose(ordered_array, transposition)

        # If spectral array, verify that length of each spectrum is equal to the number of corresponding wavelengths
        if target == 'array' and np.shape(ordered_array)[-1] != len(self.original_wavelengths):
            raise ValueError("length of wavelength dimension not equal length of original_wavelengths")

        # Load the processed array into the specified object parameter
        if target == 'array':
            self.array = ordered_array
        elif target == 'background':
            self.background = ordered_array
        else:  # This should not happen
            raise ValueError("unknown target (impossible error)")

        # If "the other array" has been loaded, warn the user if their dimensions are not compatible
        if self.array is not None and self.background is not None and \
                np.sum(np.shape(self.array)[:-1] != np.shape(self.background)) != 0:
            warnings.simplefilter('always', UserWarning)
            warnings.warn("shape of spectrum array indices does not match shape of background array")

    def load_array(self, array, names=None):
        """Load an array of spectra.

        Load `array` with dimension names `names` into the `array` parameter of the model object.

        Parameters
        ----------
        array : ndarray of ndims > 1
            An array containing at least two spectra.

        names : list of str, length = `array.ndims`
            List of dimension names for `array`. Valid dimension names are 'time', 'row', 'column' and 'wavelength'.
            'wavelength' is a required dimension.

        See Also
        --------
        load_background : Load an array of spectral backgrounds
        """
        self._load_data(array, names=names, target='array')

    def load_background(self, array, names=None):
        """Load an array of spectral backgrounds.

        Load `array` with dimension names `names` into `background` parameter of the model object.

        Parameters
        ----------
        array : ndarray of ndim>0
            An array containing at least two backgrounds.

        names : list of str, length = `array.ndims`
            List of dimension names for `array`. Valid dimension names are 'time', 'row' and 'column'.

        See Also
        --------
        load_array : Load and array of spectra
        """
        self._load_data(array, names=names, target='background')

    def train(self, X, y):
        """Fit the neural network model to spectra matrix X and spectra labels y.

        Calls the `fit` method on the `neural_network` parameter of the model object.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_spectra, n_wavelengths)
            The input spectra.
        y : ndarray of shape (n_spectra,) or (n_spectra, n_outputs)
            The target class labels.

        See Also
        --------
        test : Test how well the neural network has been trained
        """
        self.neural_network.fit(X, y)

    def test(self, X, y):
        """Test the accuracy of the trained neural network.

        Prints a table of results showing:
            1) the percentage of predictions that equal the target labels;
            2) the average classification deviation and standard deviation from the ground truth classification
                for each labelled classification;
            3) the average classification deviation and standard deviation overall.
        If the model object has an output parameter, it will create a CSV file (`self.output`/neural_network/test.csv)
        listing the predictions and ground truth data.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_spectra, n_wavelengths)
            The input spectra.
        y : ndarray of shape (n_spectra,) or (n_spectra, n_outputs)
            The target class labels.

        See Also
        --------
        train : Train the neural network
        """
        # Predict the labels
        try:
            predictions = self.neural_network.predict(X)
        except NotFittedError:
            raise NotFittedError("Neural network has not been trained yet. Call 'train' on the model class first.")
        if len(predictions) != len(y):  # Must be a corresponding ground truth label to compare to
            raise ValueError("number of samples X does not equal number of labels y")

        # Score is the percentage of predictions that are equal to their ground truth label
        score = np.sum(predictions == y) / len(predictions) * 100

        # Generate the table
        info = "+---------------------------------------------+\n"
        info += "|      Neural Network Testing Statistics      |\n"
        info += "+---------------------------------------------+\n"
        info += "| Percentage predictions==labels :: {: 6.2f}%   |\n".format(score)
        info += "+---------------------------------------------+\n"
        info += "| Average deviation for each classification   |\n"
        info += "+---------------------------------------------+\n"

        # Generate the average deviation for each classification
        classifications = np.sort(np.unique([predictions, y]))
        for classification in classifications:
            dev = predictions[np.where(y == classification)[0]] - classification
            info += "|     class {: 2d} :: {: 6.2f} ± {: 4.2f}              |\n".format(classification, np.mean(dev),
                                                                                        np.std(dev))
            info += "+---------------------------------------------+\n"

        # Generate the average deviation over all classifications
        dev = predictions - y
        info += "| Average deviation overall :: {: 6.2f} ± {: 4.2f} |\n".format(np.mean(dev), np.std(dev))
        info += "+---------------------------------------------+"

        # Print the table
        print(info)

        # Print scikit-learn classification report
        print(classification_report(y, predictions))

        # Output to file columns: 'ground_truth', 'predictions'
        if self.output is not None:
            try:
                os.makedirs(os.path.join(self.output, "neural_network"))
            except FileExistsError:
                pass
            np.savetxt(os.path.join(self.output, "neural_network", "test.csv"), np.asarray([y, predictions]).T,
                       fmt='%3d', delimiter=',', header='ground_truth, prediction')

    def _get_time_row_column(self, time=None, row=None, column=None):
        """Validate and infer the time, row and column index

        Takes any time, row and column index given and if any are not specified, they are returned as 0 if the
        spectral array only has one value at its dimension. If there are multiple and no index is specified,
        an error is raised due to the ambiguity.

        Parameters
        ----------
        time : optional, default=None
            The time index.
        row : optional, default=None
            The row index.
        column : optional, default=None
            The column index.

        Returns
        -------
        time
            The corrected time index.
        row
            The corrected row index.
        column
            The corrected column index.

        See Also
        --------
        utils.make_iter : Make a variable iterable

        Notes
        -----
        No type checking is done on the input indices so it can be anything but in most cases will need to be
        either an integer or iterable. The `utils.make_iter` function can be used to make indices iterable.
        """
        array_shape = np.shape(self.array)
        if time is None:
            # No index was specified but we can infer that it would be 0 as there is only 1 value at this dimension
            if array_shape[0] == 1:
                time = 0
            else:
                raise ValueError("time index must be specified as multiple indices exist")
        if row is None:
            if array_shape[1] == 1:
                row = 0
            else:
                raise ValueError("row index must be specified as multiple indices exist")
        if column is None:
            if array_shape[2] == 1:
                column = 0
            else:
                raise ValueError("column index must be specified as multiple indices exist")
        return time, row, column

    def get_spectra(self, time=None, row=None, column=None, spectrum=None, correct=True, background=False):
        """Gets corrected spectra from the spectral array

        Takes either a set of indices or an explicit spectrum and optionally applied corrections and background
        removal.

        Parameters
        ----------
        time : int or iterable, optional, default=None
            The time index. The index can be either a single integer index or an iterable. E.g. a list, a NumPy
            array, a Python range, etc. can be used.
        row : int or iterable, optional, default=None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default=None
            The column index. See comment for `time` parameter.
        spectrum : ndarray of ndim=1, optional, default=None
            The explicit spectrum.
        correct : bool, optional, default=True
            Whether to reinterpolate the spectrum and apply the prefilter correction (if exists).
        background : bool, optional, default=False
            Whether to include the background in the outputted spectra. Only removes the background if the
            relevant background array has been loaded. Does not remove background is processing an explicit spectrum.
        """
        # Locate the spectra
        if spectrum is None:  # No explicit spectrum so use specified indices
            time, row, column = make_iter(*self._get_time_row_column(time=time, row=row, column=column))
            spectra = self.array[time][:, row][:, :, column]  # Iterable indices are needed to crop array correctly
        else:
            if np.ndim(spectrum) != 1:  # Keep things simple (for now)
                raise ValueError("explicit spectrum must have one dimension, got %s" % np.ndim(spectrum))
            spectra = spectrum

        # Apply option corrections to spectra
        if correct:
            spectra_list = spectra.reshape(-1, spectra.shape[-1])  # To shape: (n_spectra, n_wavelengths)
            corrected_spectra = np.array([
                reinterpolate_spectrum(spectrum, self.original_wavelengths, self.constant_wavelengths)
                for spectrum in spectra_list])  # Reinterpolate each spectrum

            if self.prefilter_response is not None:  # Apply prefilter correction if one was initialised
                corrected_spectra /= self.prefilter_response

            spectra = corrected_spectra.reshape(*spectra.shape[:-1], -1)  # Revert to the original shape

        # Apply optional background removal to spectra
        if spectrum is None and not background and self.background is not None:  # Remove the background
            # Calculate transpositions to get the array in the right shape for a quick background subtraction
            forward_transpose = list(np.roll(np.arange(spectra.ndim), 1))
            backwards_transpose = list(np.roll(np.arange(spectra.ndim), -1))
            spectra = np.transpose(spectra, axes=forward_transpose) - self.background[time][:, row][:, :, column]
            spectra = np.transpose(spectra, axes=backwards_transpose)  # Revert to original shape

        return spectra

    def fit_spectrum(self, spectrum, **kwargs):
        """Fits the specified spectrum array

        Passes the spectrum argument to the fit method. For easily iterating over a list of spectra.

        Parameters
        ----------
        spectrum : ndarray of ndim=1
            The explicit spectrum.
        **kwargs : dictionary, optional
            Extra keyword arguments to pass to fit.

        Returns
        -------
        result : FitResult
            Result of the fit.

        See Also
        --------
        fit : General fitting method
        """
        return self.fit(spectrum=spectrum, **kwargs)

    def _curve_fit(self, model, spectrum, guess, sigma, bounds, x_scale, time=None, row=None, column=None):
        """scipy.optimize.curve_fit wrapper with error handling

        Passes a certain set of parameters to the scipy.optimize.curve_fit function and catches some typical
        errors, presenting a more specific warning message.

        Parameters
        ----------
        model : callable
            The model function, f(x, …). It must take the `self.constant_wavelenghts` as the first argument and the
            parameters to fit as separate remaining arguments.
        spectrum : array_like
            The dependent data, with length equal to that of `self.constant_wavelengths`.
        guess : array_like, optional
            Initial guess for the parameters to fit.
        sigma : array_like
            Determines the uncertainty in the `spectrum`. Used to weight certain regions of the spectrum.
        bounds : 2-tuple of array_like
            Lower and upper bounds on each parameter.
        x_scale : array_like
            Characteristic scale of each parameter.
        time : optional, default=None
            The time index for error handling.
        row : optional, default=None
            The row index for error handling.
        column : optional, default=None
            The column index for error handling.

        Returns
        -------
        fitted_parameters : ndarray, length=n_parameters
            The parameters that recreate the model fitted to the spectrum.
        success : bool
            Whether the fit was successful or an error had to be handled.

        See Also
        --------
        fit : General fitting method
        fit_spectrum : Explicit spectrum fitting method

        Notes
        -----
        More details can be found in the documentation for scipy.optimize.curve_fit and scipy.optimize.least_squares.
        """
        try:  # TODO Investigate if there is a performance gain to setting `check_finite` to False

            fitted_parameters = curve_fit(model, self.constant_wavelengths, spectrum,
                                          p0=guess, sigma=sigma, bounds=bounds, x_scale=x_scale)[0]
            success = True

        except RuntimeError:

            success = False
            location_text = " at ({}, {}, {})".format(time, row, column) if time is not None else ''
            warnings.warn("RuntimeError{}".format(location_text))
            fitted_parameters = np.full_like(guess, np.nan)

        except ValueError as e:

            success = False
            if str(e) != "`x0` violates bound constraints.":
                raise
            else:
                location_text = " at ({}, {}, {})".format(time, row, column) if time is not None else ''
                warnings.warn("`x0` violates bound constraints{}".format(location_text))
                fitted_parameters = np.full_like(guess, np.nan)

        except np.linalg.LinAlgError as e:

            success = False
            if str(e) != "SVD did not converge in Linear Least Squares":
                raise
            else:
                location_text = " at ({}, {}, {})".format(time, row, column) if time is not None else ''
                warnings.warn("SVD did not converge in Linear Least Squares{}.".format(location_text))
                fitted_parameters = np.full_like(guess, np.nan)

        return fitted_parameters, success



import numpy as np
from matplotlib import pyplot as plt

from mcalf.profiles.voigt import double_voigt, voigt
from mcalf.utils.spec import reinterpolate_spectrum


__all__ = ['plot_ibis8542', 'plot_spectrum']


def plot_ibis8542(wavelengths, spectrum, fit=None, background=0, sigma=None, sigma_scale=70, stationary_line_core=None,
                  subtraction=False, separate=False, output=None, figsize=None, legend_position='best', dpi=None,
                  fontfamily=None, reduced_legend=False, show_intensity=True, hook=None):
    """Plot an IBIS8542Model fit

    It is recommended to use the plot method on either an IBIS8542Model or a FitResult from an IBIS8542Model instead.

    Parameters
    ----------
    wavelengths : ndarray
        The x-axis values.
    spectrum : ndarray, length=n_wavelengths
        The y-axis values.
    fit : array_like, optional, default = None
        The fitted parameters.
    background : float or ndarray of length n_wavelengths, optional, default = 0
        The background to add to the fitted profiles.
    sigma : ndarray, length=n_wavelengths, optional, default = None
        The sigma profile used when fitting the parameters to `spectrum`. If given, will be plotted as shaded regions.
    sigma_scale : float, optional, default = 70
        A factor to multiply the error bars to change their prominence.
    stationary_line_core : float, optional, default = None
        If given, will show a dashed line at this wavelength.
    subtraction : bool, optional, default = False
        Whether to plot the `spectrum` minus emission fit (if exists) instead.
    separate : bool, optional, default = False
        Whether to plot the fitted profiles separately (if multiple components exist).
    output : str, optional, default = None
        If present, the filename to save the plot as.
    figsize : 2-tuple, optional
        Size of the figure.
    legend_position : str or int or pair of floats, optional, default = 'best'
        Position of the legend. See `matplotlib.pyplot.legend` documentation from possible values.
    dpi : int
        The number of dots per inch. For controlling the quality of the outputted figure.
    fontfamily : str, optional, default = None
        If provided, this family string will be added to the 'font' rc params group.
    reduced_legend : bool, optional, default = False
        Whether to add to the legend the labels that would be displayed on an absorption only plot. Useful for saving
        space when plotting both a single component fit and a multi-component fit alongside each other.
    show_intensity : bool, optional, default = True
        Whether to show the intensity axis tick labels and axis label.
    hook : callable, optional, default = None
        If provided this function must accept the current `plt' as a single argument such that it can operate upon
        it and make changes to the plot.

    See Also
    --------
    models.IBIS8542.plot : General plotting method
    models.IBIS8542.plot_separate : Plot the fit parameters separately
    models.IBIS8542.plot_subtraction : Plot the spectrum with the emission fit subtracted from it
    models.FitResult.plot : Plotting method on the fit result
    """
    # Choose the function to plot the fitted parameters with
    if fit is not None and len(fit) == 8:
        fit_function = double_voigt
    else:
        fit_function = voigt

    data_label = 'observation'  # Default label

    if fontfamily is not None:
        plt.rc('font', family=fontfamily)
    plt.figure(figsize=figsize, dpi=dpi)

    # If a subtraction is requested, make the relevant changes
    if subtraction and fit is not None and len(fit) == 8:
        spectrum = spectrum - voigt(wavelengths, *fit[4:], 0, clib=False)
        sigma = None
        fit = None
        data_label = 'observation - emission profile'

    # If sigma is given, make a shaded region around the spectral data.
    if sigma is not None and fit is not None:
        plt.fill_between(wavelengths, spectrum-sigma*sigma_scale, spectrum+sigma*sigma_scale, color='lightgrey')

    # Plot the spectral data
    label = None if reduced_legend else data_label
    plt.plot(wavelengths, spectrum, color='#006BA4', label=label)

    # Plot the fitted profiles if parameters are given
    if fit is not None:

        if separate and len(fit) > 4:  # Plot each component separately

            label = None if reduced_legend else 'combined profile'
            plt.plot(wavelengths, double_voigt(wavelengths, *fit, background, clib=False),
                     color='#FF800E', label=label)
            plt.plot(wavelengths, voigt(wavelengths, *fit[:4], background, clib=False),
                     color='#A2C8EC', label='absorption profile')
            plt.plot(wavelengths, voigt(wavelengths, *fit[4:], 0, clib=False),
                     color='#595959', label='emission profile')

        else:  # Plot a combined profile

            plt.plot(wavelengths, fit_function(wavelengths, *fit, background, clib=False),
                     color='#FF800E', label='fitted profile')

        label = None if reduced_legend else 'absorption line core'
        plt.axvline(x=fit[1], linestyle='--', color='#A2C8EC', label=label)
        if fit_function is double_voigt:
            plt.axvline(x=fit[5], linestyle=':', color='#595959', label='emission line core')

    if stationary_line_core is not None:  # Plot vertical dashed line
        label = None if reduced_legend else 'stationary line core'
        plt.axvline(x=stationary_line_core, linestyle='--', color='#ABABAB', label=label)

    plt.xlabel('Wavelength (Å)')
    if not reduced_legend and show_intensity:
        plt.ylabel('Intensity')
    if not show_intensity:
        plt.gca().axes.get_yaxis().set_visible(False)
        plt.gca().spines['top'].set_visible(False)
        plt.gca().spines['left'].set_visible(False)
        plt.gca().spines['right'].set_visible(False)
    plt.legend(loc=legend_position)
    plt.minorticks_on()

    if hook is not None:
        hook(plt)

    if output is not None:
        plt.savefig(output, dpi=dpi, bbox_inches='tight')

    plt.show()
    plt.close()


def plot_spectrum(wavelengths, spectrum, output=None, normalised=True, smooth=True,
                  figsize=(7, 3), dpi=600, fontfamily=None):
    """Plot a spectrum with the wavelength grid shown.

    Intended for plotting the raw data.

    Parameters
    ----------
    wavelengths : ndarray
        The x-axis values.
    spectrum : ndarray, length=n_wavelengths
        The y-axis values.
    output : str, optional, default = None
        If present, the filename to save the plot as.
    normalised : bool, optional, default = True
        Whether to normalise the spectrum using the last three spectral points.
    smooth : bool, optional, default = True
        Whether to smooth the `spectrum` with a spline.
    figsize : 2-tuple, optional, default = None
        Size of the figure.
    dpi : int, optional, default = 600
        The number of dots per inch. For controlling the quality of the outputted figure.
    fontfamily : str, optional, default = None
        If provided, this family string will be added to the 'font' rc params group.
    """

    if fontfamily is not None:
        plt.rc('font', family=fontfamily)

    dense_wavelengths = np.linspace(wavelengths[0], wavelengths[-1], num=200) if smooth else wavelengths
    dense_spectrum = reinterpolate_spectrum(spectrum, wavelengths, dense_wavelengths) if smooth else spectrum
    if normalised:
        norm = np.mean(spectrum[-3:])
        dense_spectrum /= norm
        spectrum /= norm

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    for i in range(len(wavelengths)):
        ax.axvline(x=wavelengths[i], linestyle='--', linewidth=0.5, c='black')
    ax.plot(dense_wavelengths, dense_spectrum, color='black')
    ax.plot(wavelengths, spectrum, marker='o', linestyle='None', color='black', markersize=4)

    ylabel = 'Normalised Intensity ($I/I_c$)' if normalised else 'Intensity ($I$)'
    ax.set_ylabel(ylabel)
    ax.set_xlabel('Wavelength (Å)')

    fig.subplots_adjust(bottom=0.15)

    plt.show()

    if output is not None and isinstance(output, str):
        fig.savefig(output, bbox_inches='tight', dpi=dpi)



print("import: 'mcalf'")
import mcalf

print("import: 'mcalf.models'")
import mcalf.models

print("import: 'mcalf.profiles'")
import mcalf.profiles

print("import: 'mcalf.utils'")
import mcalf.utils

print("import: 'mcalf.visualisation'")
import mcalf.visualisation




import warnings

import numpy as np
from pathos.multiprocessing import ProcessPool as Pool
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import GridSearchCV
from sklearn.neural_network import MLPClassifier

from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from mcalf.models.results import FitResult
from mcalf.models.base import ModelBase
from mcalf.profiles.voigt import voigt_nobg, double_voigt_nobg
from mcalf.utils.spec import reinterpolate_spectrum, generate_sigma
from mcalf.utils.misc import load_parameter, make_iter
from mcalf.visualisation.spec import plot_ibis8542


__all__ = ['IBIS8542Model']


class IBIS8542Model(ModelBase):
    """Class for working with IBIS 8542 Å calcium II spectral imaging observations

    Parameters
    ----------
    original_wavelengths : array_like
        One-dimensional array of wavelengths that correspond to the uncorrected spectral data.
    stationary_line_core : float, optional, default = 8542.104320687517
        Wavelength of the stationary line core.
    absorption_guess : array_like, length=4, optional, default = [-1000, stationary_line_core, 0.2, 0.1]
        Initial guess to take when fitting the absorption Voigt profile.
    emission_guess : array_like, length=4, optional, default = [1000, stationary_line_core, 0.2, 0.1]
        Initial guess to take when fitting the emission Voigt profile.
    absorption_min_bound : array_like, length=4, optional, default = [-np.inf, stationary_line_core-0.15, 1e-6, 1e-6]
        Minimum bounds for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_min_bound : array_like, length=4, optional, default = [0, stationary_line_core-0.15, 1e-6, 1e-6]
        Minimum bounds for all the emission Voigt profile parameters in order of the function's arguments.
    absorption_max_bound : array_like, length=4, optional, default = [0, stationary_line_core+0.15, 1, 1]
        Maximum bounds for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_max_bound : array_like, length=4, optional, default = [np.inf, stationary_line_core+0.15, 1, 1]
        Maximum bounds for all the emission Voigt profile parameters in order of the function's arguments.
    absorption_x_scale : array_like, length=4, optional, default = [1500, 0.2, 0.3, 0.5]
        Characteristic scale for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_x_scale : array_like, length=4, optional, default = [1500, 0.2, 0.3, 0.5]
        Characteristic scale for all the emission Voigt profile parameters in order of the function's arguments.
    neural_network : sklearn.neural_network.MLPClassifier, optional, default = see description
        The MLPClassifier object that will be used to classify the spectra. Its default value is
        `MLPClassifier(solver='lbfgs', alpha=1e-3, hidden_layer_sizes=(10, 4), random_state=1)`.
    constant_wavelengths : array_like, length same as `original_wavelengths`, optional, default = see description
        The desired set of wavelengths that the spectral data should be rescaled to represent. It is assumed
        that these have constant spacing, but that may not be a requirement if you specify your own array.
        The default value is an array from the minimum to the maximum wavelength of `original_wavelengths` in
        constant steps of `delta_lambda`, overshooting the upper bound if the maximum wavelength has not been reached.
    delta_lambda : float, optional, default = 0.05
        The step used between each value of `constant_wavelengths` when its default value has to be calculated.
    sigma : list of array_like or bool, length=(2, n_wavelengths), optional, default = [type1, type2]
        A list of different sigma that are used to weight particular wavelengths along the spectra when fitting. The
        fitting method will expect to be able to choose a sigma array from this list at a specific index. It's default
        value is `[generate_sigma(i, constant_wavelengths, stationary_line_core) for i in [1, 2]]`.
        See `utils.generate_sigma()` for more information. If bool, True will generate the default sigma value
        regardless of the value specified in `config`, and False will set `sigma` to be all ones, effectively disabling
        it.
    prefilter_response : array_like, length=n_wavelengths, optional, default = see note
        Each constant wavelength scaled spectrum will be corrected by dividing it by this array. If `prefilter_response`
        is not give, and `prefilter_ref_main` and `prefilter_ref_wvscl` are not given, `prefilter_response` will have a
        default value of `None`.
    prefilter_ref_main : array_like, optional, default = None
        If `prefilter_response` is not specified, this will be used along with `prefilter_ref_wvscl` to generate the
        default value of `prefilter_response`.
    prefilter_ref_wvscl : array_like, optional, default = None
        If `prefilter_response` is not specified, this will be used along with `prefilter_ref_main` to generate the
        default value of `prefilter_response`.
    config : str, optional, default = None
        Filename of a `.yml` file (relative to current directory) containing the initialising parameters for this
        object. Parameters provided explicitly to the object upon initialisation will override any provided in this
        file. All (or some) parameters that this object accepts can be specified in this file, except `neural_network`
        and `config`. Each line of the file should specify a different parameter and be formatted like
        `emission_guess: '[-inf, wl-0.15, 1e-6, 1e-6]'` or `original_wavelengths: 'original.fits'` for example.
        When specifying a string, use 'inf' to represent `np.inf` and 'wl' to represent `stationary_line_core` as shown.
        If the string matches a file, `utils.load_parameter()` is used to load the contents of the file.
    output : str, optional, default = None
        If the program wants to output data, it will place it relative to the location specified by this parameter.
        Some methods will only save data to a file if this parameter is not `None`. Such cases will be documented
        where relevant.

    Attributes
    ----------
    original_wavelengths : array_like
        One-dimensional array of wavelengths that correspond to the uncorrected spectral data.
    stationary_line_core : float, optional, default = 8542.099145376844
        Wavelength of the stationary line core.
    absorption_guess : array_like, length=4, optional, default = [-1000, stationary_line_core, 0.2, 0.1]
        Initial guess to take when fitting the absorption Voigt profile.
    emission_guess : array_like, length=4, optional, default = [1000, stationary_line_core, 0.2, 0.1]
        Initial guess to take when fitting the emission Voigt profile.
    absorption_min_bound : array_like, length=4, optional, default = [-np.inf, stationary_line_core-0.15, 1e-6, 1e-6]
        Minimum bounds for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_min_bound : array_like, length=4, optional, default = [0, -np.inf, 1e-6, 1e-6]
        Minimum bounds for all the emission Voigt profile parameters in order of the function's arguments.
    absorption_max_bound : array_like, length=4, optional, default = [0, stationary_line_core+0.15, 1, 1]
        Maximum bounds for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_max_bound : array_like, length=4, optional, default = [np.inf, np.inf, 1, 1]
        Maximum bounds for all the emission Voigt profile parameters in order of the function's arguments.
    absorption_x_scale : array_like, length=4, optional, default = [1500, 0.2, 0.3, 0.5]
        Characteristic scale for all the absorption Voigt profile parameters in order of the function's arguments.
    emission_x_scale : array_like, length=4, optional, default = [1500, 0.2, 0.3, 0.5]
        Characteristic scale for all the emission Voigt profile parameters in order of the function's arguments.
    neural_network : sklearn.neural_network.MLPClassifier, optional, default = see description
        The MLPClassifier object (or similar) that will be used to classify the spectra. Defaults to a `GridSearchCV`
        with `MLPClassifier(solver='lbfgs', hidden_layer_sizes=(40,), max_iter=1000)`
        for best `alpha` selected from `[1e-5, 2e-5, 3e-5, 4e-5, 5e-5, 6e-5, 7e-5, 8e-5, 9e-5]`.
    constant_wavelengths : array_like, length same as `original_wavelengths`, optional, default = see description
        The desired set of wavelengths that the spectral data should be rescaled to represent. It is assumed
        that these have constant spacing, but that may not be a requirement if you specify your own array.
        The default value is an array from the minimum to the maximum wavelength of `original_wavelengths` in
        constant steps of `delta_lambda`, overshooting the upper bound if the maximum wavelength has not been reached.
    sigma : list of array_like, length=(2, n_wavelengths), optional, default = [type1, type2]
        A list of different sigma that are used to weight particular wavelengths along the spectra when fitting. The
        fitting method will expect to be able to choose a sigma array from this list at a specific index. It's default
        value is `[generate_sigma(i, constant_wavelengths, stationary_line_core) for i in [1, 2]]`.
        See `utils.generate_sigma()` for more information.
    prefilter_response : array_like, length=n_wavelengths, optional, default = see note
        Each constant wavelength scaled spectrum will be corrected by dividing it by this array. If `prefilter_response`
        is not give, and `prefilter_ref_main` and `prefilter_ref_wvscl` are not given, `prefilter_response` will have a
        default value of `None`.
    output : str, optional, default = None
        If the program wants to output data, it will place it relative to the location specified by this parameter.
        Some methods will only save data to a file if this parameter is not `None`. Such cases will be documented
        where relevant.
    quiescent_wavelength : int, default = 1
        The index within the fitted parameters of the absorption Voigt line core wavelength.
    active_wavelength : int, default = 5
        The index within the fitted parameters of the emission Voigt line core wavelength.
    """
    def __init__(self, stationary_line_core=None,
                 absorption_guess=None, emission_guess=None,
                 absorption_min_bound=None, emission_min_bound=None,
                 absorption_max_bound=None, emission_max_bound=None,
                 absorption_x_scale=None, emission_x_scale=None,
                 neural_network=None,
                 original_wavelengths=None, constant_wavelengths=None,
                 delta_lambda=None, sigma=None, prefilter_response=None,
                 prefilter_ref_main=None, prefilter_ref_wvscl=None,
                 config=None, output=None):

        super().__init__()  # Initialise the parent class `ModelBase` also

        if config is not None:  # Process config file if one is specified

            with open(config, 'r') as stream:  # Load YAML file
                parameters = load(stream, Loader=Loader)

            # Load each parameter if it exists in the file and is not already given
            if 'stationary_line_core' in parameters and stationary_line_core is None:
                stationary_line_core = load_parameter(parameters['stationary_line_core'])
            if 'absorption_guess' in parameters and absorption_guess is None:
                absorption_guess = load_parameter(parameters['absorption_guess'], wl=stationary_line_core)
            if 'emission_guess' in parameters and emission_guess is None:
                emission_guess = load_parameter(parameters['emission_guess'], wl=stationary_line_core)
            if 'absorption_min_bound' in parameters and absorption_min_bound is None:
                absorption_min_bound = load_parameter(parameters['absorption_min_bound'], wl=stationary_line_core)
            if 'emission_min_bound' in parameters and emission_min_bound is None:
                emission_min_bound = load_parameter(parameters['emission_min_bound'], wl=stationary_line_core)
            if 'absorption_max_bound' in parameters and absorption_max_bound is None:
                absorption_max_bound = load_parameter(parameters['absorption_max_bound'], wl=stationary_line_core)
            if 'emission_max_bound' in parameters and emission_max_bound is None:
                emission_max_bound = load_parameter(parameters['emission_max_bound'], wl=stationary_line_core)
            if 'absorption_x_scale' in parameters and absorption_x_scale is None:
                absorption_x_scale = load_parameter(parameters['absorption_x_scale'])
            if 'emission_x_scale' in parameters and emission_x_scale is None:
                emission_x_scale = load_parameter(parameters['emission_x_scale'])
            if 'original_wavelengths' in parameters and original_wavelengths is None:
                original_wavelengths = load_parameter(parameters['original_wavelengths'])
            if 'constant_wavelengths' in parameters and constant_wavelengths is None:
                constant_wavelengths = load_parameter(parameters['constant_wavelengths'])
            if 'delta_lambda' in parameters and delta_lambda is None:
                delta_lambda = load_parameter(parameters['delta_lambda'])
            if 'sigma' in parameters and sigma is None:
                sigma = load_parameter(parameters['sigma'])
            if 'prefilter_response' in parameters and prefilter_response is None:
                prefilter_response = load_parameter(parameters['prefilter_response'])
            if 'prefilter_ref_main' in parameters and prefilter_ref_main is None:
                prefilter_ref_main = load_parameter(parameters['prefilter_ref_main'])
            if 'prefilter_ref_wvscl' in parameters and prefilter_ref_wvscl is None:
                prefilter_ref_wvscl = load_parameter(parameters['prefilter_ref_wvscl'])
            if 'output' in parameters and output is None:
                output = parameters['output']

        # Load default values of any parameters that haven't been given yet
        if stationary_line_core is None:
            stationary_line_core = 8542.099145376844
        if absorption_guess is None:
            absorption_guess = [-1000, stationary_line_core, 0.2, 0.1]
        if emission_guess is None:
            emission_guess = [1000, stationary_line_core, 0.2, 0.1]
        if absorption_min_bound is None:
            absorption_min_bound = [-np.inf, stationary_line_core-0.15, 1e-6, 1e-6]
        if emission_min_bound is None:
            emission_min_bound = [0, -np.inf, 1e-6, 1e-6]
        if absorption_max_bound is None:
            absorption_max_bound = [0, stationary_line_core+0.15, 1, 1]
        if emission_max_bound is None:
            emission_max_bound = [np.inf, np.inf, 1, 1]
        if absorption_x_scale is None:
            absorption_x_scale = [1500, 0.2, 0.3, 0.5]
        if emission_x_scale is None:
            emission_x_scale = [1500, 0.2, 0.3, 0.5]
        if neural_network is None:
            mlp = MLPClassifier(solver='lbfgs', hidden_layer_sizes=(40,), max_iter=1000)
            parameter_space = {'alpha': [1e-5, 2e-5, 3e-5, 4e-5, 5e-5, 6e-5, 7e-5, 8e-5, 9e-5]}  # Search region
            neural_network = GridSearchCV(mlp, parameter_space, cv=5, n_jobs=-1)  # Set GridSearchCV to find best alpha
        if original_wavelengths is None:
            raise ValueError("original_wavelengths must be specified")
        if delta_lambda is None:
            delta_lambda = 0.05
        if constant_wavelengths is None:
            constant_wavelengths = np.arange(min(original_wavelengths), max(original_wavelengths)+delta_lambda,
                                             delta_lambda)
        if prefilter_response is None:
            if prefilter_ref_main is not None and prefilter_ref_wvscl is not None:
                prefilter_response = reinterpolate_spectrum(prefilter_ref_main,
                                                            prefilter_ref_wvscl + stationary_line_core,
                                                            constant_wavelengths)
            else:
                warnings.warn("prefilter_response will not be applied to spectra")
        else:  # Make sure it is a numpy array so that division works as expected when doing array operations
            prefilter_response = np.asarray(prefilter_response, dtype=np.float64)

        # Set the object attributes (with some type enforcing)
        self.stationary_line_core = stationary_line_core
        self.absorption_guess = list(absorption_guess)
        self.emission_guess = list(emission_guess)
        self.absorption_min_bound = list(absorption_min_bound)
        self.emission_min_bound = list(emission_min_bound)
        self.absorption_max_bound = list(absorption_max_bound)
        self.emission_max_bound = list(emission_max_bound)
        self.absorption_x_scale = list(absorption_x_scale)
        self.emission_x_scale = list(emission_x_scale)
        self.neural_network = neural_network
        self.original_wavelengths = np.asarray(original_wavelengths, dtype=np.float64)
        self.constant_wavelengths = np.asarray(constant_wavelengths, dtype=np.float64)
        self.prefilter_response = prefilter_response
        self.output = output

        self.__delta_lambda = delta_lambda

        # Index of wavelength in the fitted_parameters
        self.quiescent_wavelength = 1
        self.active_wavelength = 5

        # Run some checks to make sure the specified parameters are valid
        self._validate_parameters()

        # Generate default sigma profiles for weighting fit (validate dependent parameters first)
        if sigma is None or (isinstance(sigma, bool) and sigma):
            sigma = [generate_sigma(i, self.constant_wavelengths, self.stationary_line_core) for i in [1, 2]]
        elif isinstance(sigma, bool) and not sigma:
            sigma = [np.ones(len(constant_wavelengths)) for i in [1, 2]]

        self.sigma = sigma  # Set the sigma parameter

    def _validate_parameters(self):
        """Validate some of the object's parameters

        Raises
        ------
        ValueError
            To signal that a parameter is not valid.
        """
        # Stationary line core must be a float
        if not isinstance(self.stationary_line_core, float):
            raise ValueError("stationary_line_core must be a float, got %s" % type(self.stationary_line_core))

        # Arrays must be of length 4
        if len(self.absorption_guess) != 4:
            raise ValueError("absorption_guess should be an array of length 4, got %s" %
                             len(self.absorption_guess))
        if len(self.emission_guess) != 4:
            raise ValueError("emission_guess should be an array of length 4, got %s" %
                             len(self.emission_guess))
        if len(self.absorption_min_bound) != 4:
            raise ValueError("absorption_min_bound should be an array of length 4, got %s" %
                             len(self.absorption_min_bound))
        if len(self.emission_min_bound) != 4:
            raise ValueError("emission_min_bound should be an array of length 4, got %s" %
                             len(self.emission_min_bound))
        if len(self.absorption_max_bound) != 4:
            raise ValueError("absorption_max_bound should be an array of length 4, got %s" %
                             len(self.absorption_max_bound))
        if len(self.emission_max_bound) != 4:
            raise ValueError("emission_max_bound should be an array of length 4, got %s" %
                             len(self.emission_max_bound))
        if len(self.absorption_x_scale) != 4:
            raise ValueError("absorption_x_scale should be an array of length 4, got %s" %
                             len(self.absorption_x_scale))
        if len(self.emission_x_scale) != 4:
            raise ValueError("emission_x_scale should be an array of length 4, got %s" %
                             len(self.emission_x_scale))

        # Absorption and emission fits must be constrained with the correct amplitude sign
        if self.absorption_guess[0] > 0:
            raise ValueError("absorption_guess amplitude should be negative, got %s" %
                             self.absorption_guess[0])
        if self.emission_guess[0] < 0:
            raise ValueError("emission_guess amplitude should be positive, got %s" %
                             self.emission_guess[0])
        if self.emission_min_bound[0] < 0:
            raise ValueError("emission_min_bound amplitude should be positive, got %s" %
                             self.emission_min_bound[0])
        if self.absorption_max_bound[0] > 0:
            raise ValueError("absorption_max_bound amplitude should be negative, got %s" %
                             self.absorption_max_bound[0])

        # Minimum bounds must be less that corresponding maximum bounds
        if np.sum(np.asarray(self.absorption_max_bound) - np.asarray(self.absorption_min_bound) > 0) != \
                len(self.absorption_max_bound):
            raise ValueError("values of absorption_max_bound must be greater than their "
                             "corresponding values in absorption_min_bound")
        if np.sum(np.asarray(self.emission_max_bound) - np.asarray(self.emission_min_bound) > 0) != \
                len(self.emission_max_bound):
            raise ValueError("values of emission_max_bound must be greater than their "
                             "corresponding values in emission_min_bound")

        # Wavelength arrays must be sorted ascending
        if np.sum(np.diff(self.original_wavelengths) > 0) < len(self.original_wavelengths) - 1:
            raise ValueError("original_wavelength array must be sorted ascending")
        if np.sum(np.diff(self.constant_wavelengths) > 0) < len(self.constant_wavelengths) - 1:
            raise ValueError("constant_wavelength array must be sorted ascending")

        # Warn if the constant wavelengths extrapolate the original wavelengths
        if min(self.constant_wavelengths) - min(self.original_wavelengths) < -1e-6:
            # If lower-bound of constant wavelengths is more than 1e-6 outside of the original wavelengths
            warnings.warn("Lower bound of `constant_wavelengths` is outside of `original_wavelengths` range.")
        if max(self.constant_wavelengths) - max(self.original_wavelengths) - self.__delta_lambda > 1e-6:
            # If upper-bound of constant wavelengths is more than 1e-6 ouside the original wavelengths
            warnings.warn("Upper bound of `constant_wavelengths` is outside of `original_wavelengths` range.")

        # Stationary wavelength must be within wavelength range
        original_diff = self.original_wavelengths - self.stationary_line_core
        constant_diff = self.constant_wavelengths - self.stationary_line_core
        for n, i in [['original_wavelengths', original_diff], ['constant_wavelengths', constant_diff]]:
            if min(i) > 1e-6 or max(i) < -1e-6:
                raise ValueError("`stationary_line_core` is not within `{}`".format(n))

        # If a prefilter response is given it must be a compatible length
        if self.prefilter_response is not None:
            if len(self.prefilter_response) != len(self.constant_wavelengths):
                raise ValueError("prefilter_response array must be the same length as constant_wavelengths array")

    def _get_sigma(self, classification=None, sigma=None):
        """Infer a sigma profile from the parameters provided

        If no sigma is provided, use classification to take from the model object's sigma attribute.
        If sigma is provided and is an integer, take from the model object's sigma attribute at that index,
        otherwise, return sigma as a NumPy array.

        Parameters
        ----------
        classification : int, optional, default = None
            Classification sigma profile needed to fit.
        sigma : int or array_like, optional, default = None
            Explicit sigma index or profile.

        Returns
        -------
        sigma : ndarray, length = n_constant_wavelengths
             The sigma profile.

        See Also
        --------
        utils.generate_sigma : Generate a specified sigma profile
        """
        if sigma is None:
            # Decide how to weight the wavelengths
            if classification == 0:
                return self.sigma[0]
            else:
                return self.sigma[1]
        else:
            if isinstance(sigma, int):
                return self.sigma[sigma]
            else:
                return np.asarray(sigma, dtype=np.float64)

    def classify_spectra(self, time=None, row=None, column=None, spectra=None, only_normalise=False):
        """Classify the specified spectra

        Will also normalise each spectrum such that its intensity will range from zero to one.

        Parameters
        ----------
        time : int or iterable, optional, default=None
            The time index. The index can be either a single integer index or an iterable. E.g. a list, a NumPy
            array, a Python range, etc. can be used.
        row : int or iterable, optional, default=None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default=None
            The column index. See comment for `time` parameter.
        spectra : ndarray, optional, default=None
            The explicit spectra to classify. If `only_normalise` is False, this must be 1D.
        only_normalise : bool, optional, default = False
            Whether the single spectrum given  in `spectra` should not be interpolated and corrected.

        Returns
        -------
        classifications : ndarray
            Array of classifications with the same time, row and column indices as `spectra`.

        See Also
        --------
        train : Train the neural network
        test : Test the accuracy of the neural network
        get_spectra : Get processed spectra from the objects `array` attribute
        """
        if not only_normalise:  # Get the spectrum, otherwise use the provided one directly
            spectra = self.get_spectra(time=time, row=row, column=column, spectrum=spectra)

        # Vectorised normalisation for all spectra such that intensities for each spectrum is in range [0, 1]
        spectra_list = spectra.reshape(-1, spectra.shape[-1]).T  # Reshape & transpose to (n_wavelengths, n_spectra)
        spectra_list = spectra_list - spectra_list.min(axis=0)  # Subtract each spectrum's min value from itself (min 0)
        spectra_list = spectra_list / spectra_list.max(axis=0)  # Divide each spectrum by its max value (min 0, max 1)
        spectra_list = spectra_list.T  # Transpose to (n_spectra, n_wavelengths)

        # Create the empty classifications array
        classifications = np.full(len(spectra_list), -1, dtype=int)  # -1 reserved for invalid
        valid_spectra_i = np.where(~np.isnan(spectra_list[:, 0]))  # Only predict the valid spectra
        # Note: Spectra that have indices in the data but were masked or outside the field of view should be np.nan
        try:
            classifications[valid_spectra_i] = self.neural_network.predict(spectra_list[valid_spectra_i])
        except NotFittedError:
            raise NotFittedError("Neural network has not been trained yet. Call 'train' on the model class first.")
        try:  # Try to give the classifications array the same indices as the spectral array
            classifications = classifications.reshape(*spectra.shape[:-1])
        except TypeError:  # 1D spectra array given
            assert classifications.size == 1  # Verify

        return classifications

    def _fit(self, spectrum, profile=None, sigma=None, classification=None, spectrum_index=None):
        """Fit a single spectrum for the given profile or classification

        Warning: Using this method directly will skip the corrections that are applied to spectra by the `get_spectra`
        method.

        Parameters
        ----------
        spectrum : ndarray, ndim=1, length=n_constant_wavelengths
            The spectrum to be fitted.
        profile : str, optional, default = None
            The profile to fit. (Will infer profile from classification if omitted.)
        sigma : ndarray, ndim=1, length=n_constant_wavelengths
            The sigma array with weights for each spectral wavelength.
        classification : int, optional, default = None
            Classification to determine the fitted profile to use (if profile not explicitly given).
        spectrum_index : array_like or list or tuple, length=3, optional, default = None
            The [time, row, column] index of the `spectrum` provided. Only used for error reporting.

        Returns
        -------
        result : FitResult
            Outcome of the fit returned in a FitResult object

        See Also
        --------
        fit : The recommended method for fitting spectra
        FitResult : The object that the fit method returns
        """
        if profile is None:  # If profile hasn't been specified, find it

            # Decide which profiles to fit
            if classification in [2, 3, 4]:
                profile = 'both'
            elif classification in [0, 1]:
                profile = 'absorption'
            elif classification is None:
                raise ValueError("classification must be specified if profile is not specified")
            elif not isinstance(classification, int):
                raise TypeError("classification must be an integer")
            else:
                raise ValueError("unexpected classification, got %s" % classification)

        sigma = self._get_sigma(classification, sigma)

        if profile == 'absorption':
            model = voigt_nobg
            guess = self.absorption_guess
            min_bound = self.absorption_min_bound
            max_bound = self.absorption_max_bound
            x_scale = self.absorption_x_scale
        elif profile == 'emission':
            model = voigt_nobg
            guess = self.emission_guess
            min_bound = self.emission_min_bound
            max_bound = self.emission_max_bound
            x_scale = self.emission_x_scale
        elif profile == 'both':
            model = double_voigt_nobg
            guess = self.absorption_guess + self.emission_guess
            min_bound = self.absorption_min_bound + self.emission_min_bound
            max_bound = self.absorption_max_bound + self.emission_max_bound
            x_scale = self.absorption_x_scale + self.absorption_x_scale
        else:
            raise ValueError("fit profile must be either None, 'absorption', 'emission' or 'both', got %s" % profile)

        time, row, column = spectrum_index if spectrum_index is not None else [None]*3
        fitted_parameters, success = self._curve_fit(model, spectrum, guess, sigma, (min_bound, max_bound), x_scale,
                                                     time=time, row=row, column=column)

        chi2 = np.sum(((spectrum - model(self.constant_wavelengths, *fitted_parameters)) / sigma) ** 2)

        fit_info = {'chi2': chi2, 'classification': classification, 'profile': profile,
                    'success': success, 'index': spectrum_index}

        return FitResult(fitted_parameters, fit_info)

    def fit(self, time=None, row=None, column=None, spectrum=None, profile=None, sigma=None, classifications=None,
            background=None, n_pools=None):
        """Fits the model to specified spectra

        Fits the model to an array of spectra using multiprocessing if requested.

        Parameters
        ----------
        time : int or iterable, optional, default=None
            The time index. The index can be either a single integer index or an iterable. E.g. a list, a NumPy
            array, a Python range, etc. can be used.
        row : int or iterable, optional, default=None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default=None
            The column index. See comment for `time` parameter.
        spectrum : ndarray, ndim=1, optional, default=None
            The explicit spectrum to fit the model to.
        profile : str, optional, default = None
            The profile to fit. (Will infer profile from `classifications` if omitted.)
        sigma : int or array_like, optional, default = None
            Explicit sigma index or profile. See `_get_sigma` for details.
        classifications : int, optional, default = None
            Classifications to determine the fitted profile to use (if profile not explicitly given). Will use
            neural network to classify them if not.
        background : float, optional, default = None
            If provided, this value will be subtracted from the explicit spectrum provided in `spectrum`. Will
            not be applied to spectra found from the indices, use the `load_background` method instead.
        n_pools : int, optional, default = None
            The number of processing pools to calculate the fitting over. This allocates the fitting of different
            spectra to `n_pools` separate worker processes. When processing a large number of spectra this will make
            the fitting process take less time overall. It also distributes such that each worker process has the
            same ratio of classifications to process. This should balance out the workload between workers.
            If few spectra are being fitted, performance may decrease due to the overhead associated with splitting
            the evaluation over separate processes. If `n_pools` is not an integer greater than zero, it will fit
            the spectrum with a for loop.

        Returns
        -------
        result : list of FitResult, length=n_spectra
            Outcome of the fits returned as a list of FitResult objects
        """
        # Specific fitting algorithm for IBIS Ca II 8542 Å

        # Only include the background is using an explicit spectrum; remove for indices
        include_background = False if spectrum is None else True
        spectra = self.get_spectra(time=time, row=row, column=column, spectrum=spectrum, background=include_background)

        # At least one valid spectrum must be given
        n_valid = np.sum(~np.isnan(spectra[..., 0]))
        if np.sum(~np.isnan(spectra[..., 0])) == 0:
            raise ValueError("no valid spectra given")

        if spectrum is not None and background is not None:  # remove explicit background of explicit spectrum
            spectra -= background

        if classifications is None:  # Classify if not already specified
            classifications = self.classify_spectra(spectra=spectra, only_normalise=True)

        if spectrum is None:  # No explicit spectrum given

            # Create the array of indices that the spectra represent
            time, row, column = make_iter(*self._get_time_row_column(time=time, row=row, column=column))
            indices = np.transpose(np.array(np.meshgrid(time, row, column, indexing='ij')), axes=[1, 2, 3, 0])

            # Make shape (n_spectra, n_features) so can process in a list
            spectra = spectra.reshape(-1, spectra.shape[-1])
            indices = indices.reshape(-1, indices.shape[-1])
            classifications = classifications.reshape(-1)

            # Remove spectra that are invalid (this allows for masking of the loaded data to constrain a region to fit)
            valid_spectra_i = np.where(~np.isnan(spectra[:, 0]))  # Where the first item of the spectrum is not NaN
            spectra = spectra[valid_spectra_i]
            indices = indices[valid_spectra_i]
            classifications = classifications[valid_spectra_i]

            if len(spectra) != len(indices) != len(classifications):  # Postprocessing sanity check
                raise ValueError("number of spectra, number of recorded indices and number of classifications"
                                 "are not the same (impossible error)")

            if n_pools is None or (isinstance(n_pools, int) and n_pools <= 0):  # Multiprocessing not required

                print("Processing {} spectra".format(n_valid))
                results = [self._fit(spectra[i], profile=profile, sigma=sigma, classification=classifications[i],
                                     spectrum_index=indices[i]) for i in range(len(spectra))]

            elif isinstance(n_pools, int) and n_pools >= 1:  # Use multiprocessing

                # Define single argument function that can be evaluated in the pools
                def func(data, profile=profile, sigma=sigma):
                    spectrum, index, classification = data  # Extract data and pass to `_fit` method
                    return self._fit(spectrum, profile=profile, sigma=sigma, classification=classification,
                                     spectrum_index=list(index))

                # Sort the arrays in descending classification order
                s = np.argsort(classifications)[::-1]  # Classifications indices sorted in descending order
                spectra = spectra[s]
                indices = indices[s]
                classifications = classifications[s]

                # Distribute the sorted spectra to each of `n_pools` in turn to ensure even workload
                data = []
                for pool in range(n_pools):
                    data += [[spectra[i], indices[i], classifications[i]] for i in range(pool, len(spectra), n_pools)]

                print("Processing {} spectra over {} pools".format(n_valid, n_pools))

                with Pool(n_pools) as p:  # Send the job to the `n_pools` pools
                    results = p.map(func, data)  # Map each element of `data` to the first argument of `func`
                    p.close()  # Finished, so no more jobs to come
                    p.join()  # Clean up the closed pools
                    p.clear()  # Remove the server so it can be created again

            else:
                raise TypeError("n_pools must be an integer, got %s" % type(n_pools))

        else:  # Explicit spectrum must be 1D so no loop needed
            results = self._fit(spectra, profile=profile, sigma=sigma, classification=classifications,
                                spectrum_index=None)

        return results

    def plot(self, fit=None, time=None, row=None, column=None, spectrum=None, classification=None, background=None,
             sigma=None, stationary_line_core=None, output=False, **kwargs):
        """Plots the data and fitted parameters

        Parameters
        ----------
        fit : FitResult or list or array_like, optional, default = None
            The fitted parameters to plot with the data. Can extract the necessary plot metadata from the fit object.
            Otherwise, `fit` should be the parameters to be fitted to either a Voigt or double Voigt profile depending
            on the number of parameters fitted.
        time : int or iterable, optional, default = None
            The time index. The index can be either a single integer index or an iterable. E.g. a list, a NumPy
            array, a Python range, etc. can be used.
        row : int or iterable, optional, default = None
            The row index. See comment for `time` parameter.
        column : int or iterable, optional, default = None
            The column index. See comment for `time` parameter.
        spectrum : ndarray of length `original_wavelengths`, ndim=1, optional, default = None
            The explicit spectrum to plot along with a fit (if specified).
        classification : int, optional, default = None
            Used to determine which sigma profile to use. See `_get_sigma` for more details.
        background : float or array_like of length `constant_wavelengths`, optional, default = see note
            Background to added to the fitted profiles. If a `spectrum` is given, this will default to zero, otherwise
            the value loaded by `load_background` will be used.
        sigma : int or array_like, optional, default = None
            Explicit sigma index or profile. See `_get_sigma` for details.
        stationary_line_core : float, optional, default = `self.stationary_line_core`
            The stationary line core wavelength to mark on the plot.
        output : bool or str, optional, default = False
            Whether to save the plot to a file. If true, a file of format `plot_<time>_<row>_<column>.eps` will be
            created in the current directory. If a string, that will be used as the filename. (Can change filetype
            like this.) If false, no file will be created.
        **kwargs
            Parameters used by matplotlib and `separate` (see `plot_separate`) and `subtraction`
            (see `plot_subtraction`).
                - `figsize` passed to `matplotlib.pyplot.figure`
                - `legend_position` passed to `matplotlib.pyplot.legend`
                - `dpi` passed to `matplotlib.pyplot.figure` and `matplotlib.pyplot.savefig`
                - `fontfamily` passed to `matplotlib.pyplot.rc('font', family=`fontfamily`)` if given

        See Also
        --------
        plot_separate : Plot the fit parameters separately
        plot_subtraction : Plot the spectrum with the emission fit subtracted from it
        FitResult.plot : Plotting method on the fit result
        """
        if fit.__class__ == FitResult:  # If fit is a `FitResult`

            # Extract classification (so it can choose the correct sigma to plot)
            if classification is None:
                classification = fit.classification

            # Extract spectrum index (so it can load the raw data)
            if (time, row, column) == (None, None, None) and fit.index is not None:
                time, row, column = fit.index

            # Replace fit with just the fitted parameters
            fit = fit.parameters

        # Has a spectrum been given? (used for tidying plot later)
        explicit_spectrum = False if spectrum is None else True

        # Get the spectrum and reduce it to 1D
        spectrum = self.get_spectra(time=time, row=row, column=column, spectrum=spectrum, background=True)
        spectrum = np.squeeze(spectrum)
        if spectrum.ndim != 1:
            raise ValueError("spectrum must be 1D")

        # If no background to be added to the fit is given...
        if background is None and fit is not None:
            if not explicit_spectrum:  # Take from loaded background if using indices
                time, row, column = self._get_time_row_column(time=time, row=row, column=column)
                if sum([isinstance(i, (int, np.int64, np.int32, np.int16, np.int8)) for i in [time, row, column]]) != 3:
                    raise TypeError("plot only accepts integer values for time, row and column")
                background = self.background[time, row, column]
            else:  # Otherwise assume to be zero
                background = 0

        if classification is not None or sigma is not None:  # Skip if neither classification or sigma are given
            sigma = self._get_sigma(classification, sigma)

        if stationary_line_core is None:
            stationary_line_core = self.stationary_line_core

        if isinstance(output, bool):
            if output:
                # Create a name
                output = "plot_{}_{}_{}.eps".format(time, row, column)
            else:
                output = None  # Do not output a file
        elif not isinstance(output, str):  # A valid string can pass through as the filename
            raise TypeError("output must be either boolean or a string, got %s" % type(output))

        plot_ibis8542(self.constant_wavelengths, spectrum, fit=fit, background=background, sigma=sigma,
                      stationary_line_core=stationary_line_core, output=output, **kwargs)

    def plot_separate(self, *args, **kwargs):
        """Plot the fitted profiles separately

        If multiple profiles exist, fit them separately. See `plot` for more details.

        See Also
        --------
        plot : General plotting method
        plot_subtraction : Plot the spectrum with the emission fit subtracted from it
        FitResult.plot : Plotting method on the fit result
        """
        self.plot(*args, separate=True, **kwargs)

    def plot_subtraction(self, *args, **kwargs):
        """Plot the spectrum with the emission fit subtracted from it

        If multiple profiles exist, subtract the fitted emission from the raw data. See `plot` for more details.

        See Also
        --------
        plot : General plotting method
        plot_separate : Plot the fit parameters separately
        FitResult.plot : Plotting method on the fit result
        """
        self.plot(*args, subtraction=True, **kwargs)



import glob

import numpy as np
from matplotlib import pyplot as plt, colors, cm
from matplotlib.gridspec import GridSpec


__all__ = ['plot_classifications', 'plot_class_map', 'plot_averaged_class_map']


def plot_classifications(class_map, spectra, labels, extent=(0, 200, 0, 200), xticks=(0, 15, 3), yticks=(0, 15, 3),
                         xscale=0.725*0.097, yscale=0.725*0.097, output=None, figsize=None, dpi=600, fontfamily=None):
    """Plot the spectra separated into their classifications along with an example classified map.

    Must be 5 classifications.

    Parameters
    ----------
    class_map : ndarray, ndim=2
        Two-dimensional array of classifications.
    spectra : ndarray, ndim=2
        Two-dimensional array with dimensions [spectra, wavelengths].
    labels : ndarray, ndim=1, length of `spectra`
        List of classifications for each spectrum in `spectra`.
    output : str, optional, default = None
        If present, the filename to save the plot as.
    figsize : 2-tuple, optional, default = None
        Size of the figure.
    dpi : int, optional, default = 600
        The number of dots per inch. For controlling the quality of the outputted figure.
    fontfamily : str, optional, default = None
        If provided, this family string will be added to the 'font' rc params group.
    vmin : float, optional, default = -max(|`velmap`|)
        Minimum velocity to plot. If not given, will be -vmax, for vmax not None.
    vmax : float, optional, default = max(|`velmap`|)
        Maximum velocity to plot. If not given, will be -vmin, for vmin not None.
    extent : 4-tuple, optional, default = (0, 200, 0, 200)
        Region the `velmap` is cropped to.
    xticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the x-axis ticks in Mm.
    yticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the y-axis ticks in Mm.
    xscale : float, optional = 0.725 * 0.097
        Scaling factor between x-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    yscale : float, optional = 0.725 * 0.097
        Scaling factor between y-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    """

    if fontfamily is not None:
        plt.rc('font', family=fontfamily)
    fig = plt.figure(constrained_layout=True, figsize=figsize, dpi=dpi)

    gs = GridSpec(2, 3, figure=fig, wspace=0)
    ax1 = fig.add_subplot(gs[0, 0])
    ax2 = fig.add_subplot(gs[0, 1])
    ax3 = fig.add_subplot(gs[0, 2])
    ax4 = fig.add_subplot(gs[1, 0])
    ax5 = fig.add_subplot(gs[1, 1])
    map_plot = fig.add_subplot(gs[1, 2])
    axes = [ax1, ax2, ax3, ax4, ax5, map_plot]

    # Optimised for readers with color blindness
    cmap = colors.ListedColormap(['#0072b2', '#56b4e9', '#009e73', '#e69f00', '#d55e00'])

    for classification in range(5):
        n_plots = 0  # Number plotted for this classification
        ax = axes[classification]  # Select the axis
        class_colour = cmap(classification)
        for j in range(len(labels)):
            if labels[j] == classification:
                n_plots += 1
                ax.plot(spectra[j], linewidth=0.5, color=class_colour)
                if n_plots >= 20:  # Only plot the first 20 of each classification
                    break
        ax.set_xticks([])  # No wavelengths plotted
        ax.set_yticks([0, 1])  # Only show that intensity is scaled [0, 1]
        ax.margins(0)

    ax = axes[-1]  # Classification map will be placed in the last axis

    cmap.set_bad(color='white')  # Background for masked points
    class_map_float = np.asarray(class_map, dtype=float)
    class_map_float[class_map == -1] = np.nan
    classif_img = ax.imshow(class_map_float[::-1], cmap=cmap, vmin=-0.5, vmax=4.5, interpolation='nearest')

    ax.set_xlim(*extent[:2]), ax.set_ylim(*extent[2:])

    xticks_Mm = np.arange(*xticks)
    xticks = (xticks_Mm / xscale) + extent[0]
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticks_Mm)
    ax.set_xlabel('Distance (Mm)')

    yticks_Mm = np.arange(*yticks)
    yticks = (yticks_Mm / yscale) + extent[2]
    ax.set_yticks(yticks)
    ax.set_yticklabels(yticks_Mm)
    ax.set_ylabel('Distance (Mm)')

    cbar = fig.colorbar(classif_img, ax=axes, ticks=[0, 1, 2, 3, 4], orientation='horizontal', shrink=1, pad=0)
    cbar.ax.set_xticklabels(['0\nabsorption', '1', '2', '3', '4\nemission'])

    plt.show()

    if output is not None and isinstance(output, str):
        fig.savefig(output, bbox_inches='tight', dpi=dpi)


def plot_class_map(class_map, overall_classes=None, classes=None, time_index=None, cadence=None,
                   xticks=(0, 15, 2), yticks=(0, 15, 2), xscale=0.725 * 0.097, yscale=0.725 * 0.097,
                   output=None, file_prefix='classmap_plot_', file_ext='png',
                   figsize=(5 * 3 / 2.5, 3 * 3 / 2.5), dpi=600, fontfamily=None, cache=False):
    """Plot an image of the classifications at a particular time along with bar charts of the classifications

    Parameters
    ----------
    class_map : ndarray, ndim=2 or 3
        Two-dimensional array of classifications. If three dimensions are given, the first dimension is assumed to
        represent the time.
    overall_classes : ndarray or bool, optional
        The percentage of spectra that belong to each classification in the overall dataset. If omitted, these will
        be calculated used all of the classifications given is `class_map`. If true is given, these will also be
        calculated in the same way and returned without any plotting done. (This returned array can then be used to
        speed up later calls of this function.)
    classes : ndarray, optional, default = ndarray of [0, 1, 2, 3, 4]
        Array of all the possible classifications in `class_map`.
    time_index : int, optional, default = 0
        The index of the time dimension of `class_map`, required if class_map is 3D. Also used for plotting the time.
    cadence : float, units = seconds, optional, default = None
        If given, the time index will be multiplied by this value and converted into a time in minutes on the plot.
        Otherwise, the `time_index` will be plotted without units.
    xticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the x-axis ticks in Mm.
    yticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the y-axis ticks in Mm.
    xscale : float, optional = 0.725 * 0.097
        Scaling factor between x-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    yscale : float, optional = 0.725 * 0.097
        Scaling factor between y-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    output : str or bool, optional, default = None
        If present, the filename to save the plot as. If omitted, the plot will not be saved. If true, the filename
        will be generated using the `time_index` along with the `file_prefix` and `file_ext`.
    file_prefix : str, optional, default = 'classmap_plot_'
        The prefix to use in the filename when `output` is true.
    file_ext : str, optional, default = 'png'
        The file extension (without the dot) to use when `output` is true.
    figsize : 2-tuple, optional, default = None
        Size of the figure.
    dpi : int, optional, default = 600
        The number of dots per inch. For controlling the quality of the outputted figure.
    fontfamily : str, optional, default = None
        If provided, this family string will be added to the 'font' rc params group.
    cache : bool, optional, default = False
        If true, the plot will not be regenerated if the output filename already exists.

    Returns
    -------
    overall_classes : ndarray
        If `overall_classes` is initially true, their calculated values will be returned.
    """
    if classes is None:
        classes = np.arange(5, dtype=int)

    if overall_classes is None or isinstance(overall_classes, bool):
        just_print_overall_classes = True if overall_classes else False
        overall_classes = class_map.flatten()
        counts = np.zeros(len(classes))
        for i in classes:
            counts[i] = len(overall_classes[overall_classes == i])
        overall_classes = counts / len(overall_classes) * 100  # Convert to percentage
        if just_print_overall_classes:
            return overall_classes

    if class_map.ndim == 3:
        if time_index is None:
            raise ValueError('A `time_index` must be specified as multiple time dimensions are in `class_map`.')
        class_map = class_map[time_index]
    else:
        if class_map.ndim != 2:
            raise ValueError('`class_map` must have either 2 or 3 dimensions, got %s' % class_map.ndim)
        if time_index is None:
            time_index = 0

    if isinstance(output, bool) and output:
        output = '{}{:05d}.{}'.format(file_prefix, time_index, file_ext)

    if cache and output is not None and len(glob.glob(output)) > 0:
        return 0

    if fontfamily is not None:
        plt.rc('font', family=fontfamily)

    time = time_index if cadence is None else time_index * cadence / 60
    time_unit = '' if cadence is None else ' min'
    time_prefix = 't = ' if cadence is None else ''

    cmap_colors = np.array(['#0072b2', '#56b4e9', '#009e73', '#e69f00', '#d55e00'])[:len(classes)]
    cmap = colors.ListedColormap(cmap_colors)
    extent = (0, len(class_map[0]), 0, len(class_map))
    cmap.set_bad(color='#999999', alpha=1)
    bar_colors = cmap(classes)

    fig = plt.figure(figsize=figsize, dpi=dpi, constrained_layout=True)
    gs = GridSpec(2, 3, figure=fig)
    ax1 = fig.add_subplot(gs[:, :2])
    ax2 = fig.add_subplot(gs[0, 2])
    ax3 = fig.add_subplot(gs[1, 2])

    plt.sca(ax1)
    im = plt.imshow(class_map, cmap=cmap, vmin=min(classes)-0.5, vmax=max(classes)+0.5, extent=extent,
                    interpolation='nearest')
    fig.colorbar(im, ax=ax1, ticks=classes, orientation='vertical', label='absorption' + ' ' * 41 + 'emission')
    ax1.set_title('Classifications at {}{:.2f}{}'.format(time_prefix, time, time_unit))

    xticks_Mm = np.arange(*xticks)
    xticks = (xticks_Mm / xscale) + extent[0]
    ax1.set_xticks(xticks)
    ax1.set_xticklabels(xticks_Mm)
    ax1.set_xlabel('Distance (Mm)')

    yticks_Mm = np.arange(*yticks)
    yticks = (yticks_Mm / yscale) + extent[2]
    ax1.set_yticks(yticks)
    ax1.set_yticklabels(yticks_Mm)
    ax1.set_ylabel('Distance (Mm)')

    current_classes = class_map.flatten()
    counts = np.zeros(len(classes))
    for i in classes:
        counts[i] = len(current_classes[current_classes == i])
    current_classes = counts / len(current_classes) * 100  # Convert to percentage

    plt.sca(ax2)
    plt.bar(classes, current_classes, color=bar_colors)
    ax2.set_title('Current Classes (%)')

    plt.sca(ax3)
    plt.bar(classes, overall_classes, color=bar_colors)
    ax3.set_title('Overall Classes (%)')

    for ax in [ax2, ax3]:
        ax.set_xlabel(None)
        ax.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
        ax.set_yscale('log')
        ax.set_ylim(0.01, 100)

    plt.show()

    if output is not None and isinstance(output, str):
        fig.savefig(output, bbox_inches='tight', dpi=dpi)


def plot_averaged_class_map(class_map, classes=None, continuous=False,
                   xticks=(0, 15, 2), yticks=(0, 15, 2), xscale=0.725 * 0.097, yscale=0.725 * 0.097,
                   output=None, figsize=None, dpi=600, fontfamily=None):
    """Plot an image of the time averaged classifications

    Parameters
    ----------
    class_map : ndarray, ndim=3
        Three-dimensional array of classifications, with the times given in the first dimension.
    classes : ndarray, optional, default = ndarray of [0, 1, 2, 3, 4]
        Array of all the possible classifications in `class_map`.
    continuous : bool, optional, default = False
        Whether to plot the with a continuous color scale or round to the nearest classification.
    xticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the x-axis ticks in Mm.
    yticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the y-axis ticks in Mm.
    xscale : float, optional = 0.725 * 0.097
        Scaling factor between x-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    yscale : float, optional = 0.725 * 0.097
        Scaling factor between y-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    output : str, optional, default = None
        If present, the filename to save the plot as. If omitted, the plot will not be saved.
    figsize : 2-tuple, optional, default = None
        Size of the figure.
    dpi : int, optional, default = 600
        The number of dots per inch. For controlling the quality of the outputted figure.
    fontfamily : str, optional, default = None
        If provided, this family string will be added to the 'font' rc params group.
    """

    if classes is None:
        classes = np.arange(5, dtype=int)

    if class_map.ndim != 3:
        raise ValueError('`class_map` must have 3 dimensions, got %s' % class_map.ndim)

    class_map = np.mean(class_map, axis=0)

    if fontfamily is not None:
        plt.rc('font', family=fontfamily)

    if continuous:
        cmap = cm.get_cmap('binary_r')
        vmin = min(classes)
        vmax = max(classes)
    else:
        cmap_colors = np.array(['#0072b2', '#56b4e9', '#009e73', '#e69f00', '#d55e00'])[:len(classes)]
        cmap = colors.ListedColormap(cmap_colors)
        vmin = min(classes) - 0.5
        vmax = max(classes) + 0.5

    extent = (0, len(class_map[0]), 0, len(class_map))
    cmap.set_bad(color='#999999', alpha=1)

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi, constrained_layout=True)

    im = ax.imshow(class_map, cmap=cmap, vmin=vmin, vmax=vmax, extent=extent, interpolation='nearest')
    fig.colorbar(im, ax=ax, ticks=classes, orientation='vertical', label='absorption' + ' ' * 47 + 'emission')

    xticks_Mm = np.arange(*xticks)
    xticks = (xticks_Mm / xscale) + extent[0]
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticks_Mm)
    ax.set_xlabel('Distance (Mm)')

    yticks_Mm = np.arange(*yticks)
    yticks = (yticks_Mm / yscale) + extent[2]
    ax.set_yticks(yticks)
    ax.set_yticklabels(yticks_Mm)
    ax.set_ylabel('Distance (Mm)')

    plt.show()

    if output is not None and isinstance(output, str):
        fig.savefig(output, bbox_inches='tight', dpi=dpi)



import numpy as np
from matplotlib import pyplot as plt


__all__ = ['plot_map']


def plot_map(velmap, mask=None, umbra_mask=None, figsize=None, vmin=None, vmax=None, extent=None,
             xticks=(0, 15, 2), yticks=(0, 15, 2), xscale=0.725 * 0.097, yscale=0.725 * 0.097,
             output=None, dpi=600, fontfamily=None, units="km/s", linewidths=None):
    """Plot the velocity map

    Plots a velocity map for publication.

    Parameters
    ----------
    velmap : ndarray, ndim=2
        Two-dimensional array of velocities.
    mask : ndarray of bool, ndim=2, shape `velmap`, optional, default = None
        Mask showing the region where velocities were found for. True is outside the
        velocity region and False is where valid velocities should be found. Specifying
        a mask allows for errors in the velocity calculation to be black and points
        outside the region to be gray. If omitted, all invalid points will be gray.
    umbra_mask : ndarray of bool, ndim=2, shape `velmap`, optional, default = None
        A mask of the umbra, True outside, False inside. If given, a contour will
        outline the umbra, or other feature the mask represents.
    output : str, optional, default = None
        If present, the filename to save the plot as.
    figsize : 2-tuple
        Size of the figure.
    dpi : int
        The number of dots per inch. For controlling the quality of the outputted figure.
    fontfamily : str, optional, default = None
        If provided, this family string will be added to the 'font' rc params group.
    vmin : float, optional, default = -max(|`velmap`|)
        Minimum velocity to plot. If not given, will be -vmax, for vmax not None.
    vmax : float, optional, default = max(|`velmap`|)
        Maximum velocity to plot. If not given, will be -vmin, for vmin not None.
    extent : 4-tuple, optional, default = (0, n_rows, 0, n_columns)
        Data range of `velmap`. TODO: Remove (assume one-to-one relationship)
    xticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the x-axis ticks in Mm.
    yticks : 3-tuple, optional, default = (0, 15, 2)
        The start, stop and step for the y-axis ticks in Mm.
    xscale : float, optional = 0.725 * 0.097
        Scaling factor between x-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    yscale : float, optional = 0.725 * 0.097
        Scaling factor between y-axis data coordinate steps and 1 Mm. Mm = data / xscale.
    units : str, optional, default = 'km/s'
        The units of `velmap` data. Printed on colorbar.
    linewidths : float or sequence of floats, optional, default = None
        The width of the contours plotted for `umbra_mask`.
    """
    if fontfamily is not None:
        plt.rc('font', family=fontfamily)
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    cmap = plt.get_cmap('bwr')

    velmap_cropped = velmap.copy()

    if extent is None:
        extent = (0, len(velmap_cropped), 0, len(velmap_cropped[0]))

    if mask is not None:
        # Show invalid pixels outside the mask as black, inside as gray
        if mask.shape != velmap_cropped.shape:
            raise ValueError("`mask` must be the same shape as `velmap`")
        unmasked_section = np.empty_like(mask, dtype=float)
        unmasked_section[mask] = np.nan
        unmasked_section[~mask] = 1
        cmap_mask = plt.get_cmap('gray')
        cmap_mask.set_bad(color='#999999', alpha=1)
        ax.imshow(unmasked_section, cmap=cmap_mask, extent=extent, interpolation='nearest')
        cmap.set_bad(color='#000000', alpha=0)
        velmap_cropped[mask] = np.nan
    else:
        cmap.set_bad(color='#999999', alpha=1)

    # Show the velocities
    if vmin is None and vmax is None:
        vmax = np.nanmax(np.abs(velmap))
        vmin = -vmax
    elif vmin is None:
        vmin = -vmax
    elif vmax is None:
        vmax = -vmin
    im = ax.imshow(velmap_cropped, cmap=cmap, vmin=vmin, vmax=vmax, extent=extent, interpolation='nearest')

    if umbra_mask is not None:
        if umbra_mask.shape != velmap_cropped.shape:
            raise ValueError("`umbra_mask` must be the same shape as `velmap`")
        plt.contour(umbra_mask[::-1], cmap='binary', extent=extent, linewidths=linewidths)

    plt.colorbar(im, ax=ax, label='Doppler velocity ({})'.format(units))

    xticks_Mm = np.arange(*xticks)
    xticks = (xticks_Mm / xscale) + extent[0]
    plt.xticks(xticks, xticks_Mm)
    plt.xlabel('Distance (Mm)')

    yticks_Mm = np.arange(*yticks)
    yticks = (yticks_Mm / yscale) + extent[2]
    plt.yticks(yticks, yticks_Mm)
    plt.ylabel('Distance (Mm)')

    plt.show()

    if output is not None and isinstance(output, str):
        fig.savefig(output, bbox_inches='tight', dpi=dpi)
